import numpy as np

from typing import Iterable, Dict
from scipy import interpolate
from abc import ABCMeta, abstractmethod

from .. params import parameters
from .. params.parameters import BinaryInitialParameters
from .. tools.utils import (
    radialcurves_mean_error,
    lightcurves_mean_error,
    time_layer_resolver
)
from .. models import lc as lc_model
from .. models import rv as rv_model
from ... utils import is_empty
from ... import units
from ... import settings
from ... observer.observer import Observer
from ... observer.utils import normalize_light_curve
from ... binary_system.system import BinarySystem
from ... binary_system.curves.community import RadialVelocitySystem


from ... logger import getLogger

logger = getLogger("analytics.binary_fit.shared")


class AbstractFit(metaclass=ABCMeta):
    """
    General framework for solution of the inverse problem in ELISa.

    Slotted attributes:

        :constrained: Dict; list of constrained model parameters in flat JSON format
        :discretization: int; discretization factor for primary component
        :fitable: Dict; list of optimized model parameters in flat JSON format
        :fixed: Dict; list of constant model parameters in flat JSON format
        :initial_vector: List; array of starting values for variable parameters
        :interp_treshold: int; Above this total number of datapoints, curve will be interpolated
                               using model containing `interp_treshold` equidistant points per epoch.
        :normalization: Dict[str, tuple]; normalization boundaries of variable parameters
                                          {parameter@name: (min, max), ...}
        :num_of_points: Dict[str, int]; number of points of observations in each passband
        :observer: elisa.observer.observer.Observer; Observer instance
        :x_data: Dict[str, numpy.array]; phases or times of observations in each filter
        :x_data_reduced: numpy.array; phases or times of observations grouped from all filters without duplicate phases
        :x_data_reducer: Dict[str, numpy.array]; x_data[passband] = x_data_reduced[x_data_reducer[passband]]
        :y_data: Dict[str, numpy.array]; fluxes in each filter
        :y_err: Dict[str, numpy.array]; observational errors of y_data
        :atmosphere_model: Union[dict(), None]: atmosphere models for each component {component: atm_model, ...}
        :limb_darkening_coefficients: Union[dict(), None]: custom limb-darkening coefficients used for components
                                      in format {component: {passband: ld_coeffs, }, }
    """
    MEAN_ERROR_FN = None

    __slots__ = ['fixed', 'constrained', 'fitable', 'observer', 'x_data', 'y_data', 'num_of_points'
                 'y_err', 'x_data_reduced', 'x_data_reducer', 'initial_vector', 'normalization', 'flat_result',
                 'discretization', 'interp_treshold', 'atmosphere_model', 'limb_darkening_coefficients']

    def set_up(self, x0: BinaryInitialParameters, data: Dict, passband: Iterable = None, **kwargs):
        """
        Setting up the class attributes listed in __slots__.

        :param x0: BinaryInitialParameters; initial state of model parameters
        :param data: Dict[Union[elisa.analytics.dataset.base.LCData, elisa.analytics.dataset.base.RVData]];
                     observational data in photometric filters
        :param passband: List; list of used passbands, use None in case of RV fit
        :param kwargs: Dict; class dependent content (see inheritor classes)
        """
        self._check_data_param_consistency(x0, data)
        setattr(self, 'fixed', x0.get_fixed(jsonify=False))
        setattr(self, 'constrained', x0.get_constrained(jsonify=False))
        setattr(self, 'fitable', x0.get_fitable(jsonify=False))
        setattr(self, 'normalization', x0.get_normalization_map())

        observer = Observer(passband='bolometric' if passband is None else passband, system=None)
        setattr(self, 'observer', observer)
        setattr(self.observer, 'system_cls', kwargs.get('observer_system_cls'))

        setattr(self, 'x_data', {key: val.x_data for key, val in data.items()})
        setattr(self, 'y_data', {key: val.y_data for key, val in data.items()})
        setattr(self, 'num_of_points', {key: np.shape(val.y_data)[0] for key, val in data.items()})

        err = {key: abs(self.__class__.MEAN_ERROR_FN(val)) if data[key].y_err is None else data[key].y_err
               for key, val in self.y_data.items()}
        setattr(self, 'y_err', err)

        x_data_reduced, x_data_reducer = parameters.xs_reducer({key: val.x_data for key, val in data.items()})
        setattr(self, 'x_data_reduced', x_data_reduced)
        setattr(self, 'x_data_reducer', x_data_reducer)

        setattr(self, 'initial_vector', [val.value for val in self.fitable.values()])
        setattr(self, 'flat_result', dict())

        if not isinstance(getattr(self, 'atmosphere_model', None), (dict, type(None))):
            raise ValueError('Please provide argument `atmosphere_mode` in format '
                             '{component (primary or secondary): atmosphere_model_name (ck04, bb, ...), }')

        if not isinstance(getattr(self, 'limb_darkening_coefficients', None), (dict, type(None))):
            raise ValueError('Please provide argument `limb_darkening_coefficients` in format '
                             '{component (primary or secondary): {passband: ld_coeffs (float, list), }, }')

    @staticmethod
    def _check_data_param_consistency(x0: BinaryInitialParameters, data: Dict):
        """
        Checking whether input observations have time stamps either in JD or photometric phase.
        Function also cheks for compatibility between input observational data and initial parameters.

        :param x0: BinaryInitialParameters; initial state of model parameters
        :param data: Dict[Union[elisa.analytics.dataset.base.LCData, elisa.analytics.dataset.base.RVData]];
                     observational data in photometric filters:
        :return: None
        """
        x_units = {key: val.x_unit for key, val in data.items()}
        x_units_reduced = list(set(x_units.values()))
        if len(x_units_reduced) > 1:
            raise ValueError(f"Please make sure that all synthetic observations are supplied "
                             f"in BJD or photometric phases. Current state: {x_units}")

        try:
            getattr(x0, 'primary_minimum_time')
        except AttributeError:
            if x_units_reduced[0] == units.d:
                raise ValueError("Your initial parameters do not contain `primary_minimum_time`, yet you provided "
                                 "your synthetic observations in JD. Either convert your observations to photometric "
                                 "phases using `DataSet.convert_to_phases(period, t0)` or include "
                                 "`primary_minimum_time` to your initial fit parameters.")
        else:
            if x_units_reduced[0] == units.dimensionless_unscaled:
                raise ValueError("Your initial parameters contain `primary_minimum_time`, yet you provided "
                                 "your synthetic observations in photometric phases. Either convert your observations "
                                 "to JD using `DataSet.convert_to_time(period, t0)` or remove `primary_minimum_"
                                 "time` from your initial fit parameters.")

    @abstractmethod
    def fit(self, *args, **kwargs):
        pass

    # noinspection PyUnresolvedReferences
    @staticmethod
    def eval_constrained_results(result_dict: Dict[str, Dict], constraints: Dict[str, 'InitialParameter']):
        """
        Function adds values of constrained parameters into the resulting dictionary based on the values of the
        dependent variables.

        :param constraints: Dict; contains constrained parameters
        :param result_dict: Dict; {'name': {'value': value, 'unit': unit, ...}
        :return: Dict; {'name': {'value': value, 'unit': unit, ...} JSON with evaluated constraints
        """
        if is_empty(constraints):
            return result_dict

        res_val_dict = {key: val['value'] for key, val in result_dict.items()}
        constrained_values = parameters.constraints_evaluator(res_val_dict, constraints)
        result_dict.update(
            {
                key: {
                    'value': val,
                    'constraint': constraints[key].constraint,
                    'unit': constraints[key].to_dict()['unit']
                } for key, val in constrained_values.items()
            })
        return result_dict


class AbstractRVFit(AbstractFit):
    """
    Abstract implementation of the RV fit.
    """
    MEAN_ERROR_FN = radialcurves_mean_error

    __slots__ = []  # inheriting attributes from parent

    @abstractmethod
    def fit(self, *args, **kwargs):
        pass

    def set_up(self, x0: BinaryInitialParameters, data: Dict, **kwargs):
        """
        Setting up class attributes inherited in __slots__.

        :param x0: BinaryInitialParameters; initial state of model parameters,
        :param data: Dict[str, elisa.analytics.dataset.base.RVData]; RV data for primary and secondary component,
        :param kwargs: Dict;
        :**kwargs options**:
            * :observer_system_cls: -- Union[BinarySystem, RadialVelocitySystem]; system used to evaluate synthetic
                                                                                  observations
        :return:
        """
        super().set_up(x0, data, passband=None, **kwargs)

    # noinspection PyUnusedLocal
    def coefficient_of_determination(self, model_parameters, data, discretization, interp_treshold):
        """
        Function returns R^2 for given model parameters and observed data.

        :param model_parameters: Dict; set of model parameters in JSON format
        :param data: Dict[str, RVData]; Dict[str, RVData]; observed RVs for each component
        :param discretization: not (yet) used
        :param interp_treshold: not (yet) used
        :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
        """
        self.set_up(parameters.BinaryInitialParameters(**model_parameters), data,
                    observer_system_cls=RadialVelocitySystem)
        r_squared_args = self.x_data_reduced, self.y_data, self.x_data_reducer, self.observer.system_cls
        flat_result = parameters.deserialize_result(model_parameters)
        r_dict = {key: value['value'] for key, value in flat_result.items()}

        logger.info("Evaluating light curve for calculation of R^2.")
        r_squared_result = rv_r_squared(rv_model.central_rv_synthetic, *r_squared_args, **r_dict)
        logger.info("Calculation of R^2 finished.")
        return r_squared_result


class AbstractLCFit(AbstractFit):
    """
    Abstract implementation the LC fitting.
    """
    MEAN_ERROR_FN = lightcurves_mean_error

    __slots__ = []  # inheriting attributes from parent

    def set_up(self, x0: BinaryInitialParameters, data: Dict, passband: Iterable = None, **kwargs):
        """
        Setting up class attributes inherited in __slots__.

        :param x0: BinaryInitialParameters; initial state of model parameters,
        :param data: Dict[str, elisa.analytics.dataset.base.LCData]; observations in each filter,
        :param passband: List; list of used passbands
        :param kwargs: Dict;
        :**possible kwargs**:
            * :observer_system_cls: Union[BinarySystem, RadialVelocitySystem]; system used to evaluate synthetic
                observations
            * :discretization: float; discretization factor for the primary component;
            * :samples: Union[str, List]; `uniform`, `adaptive` or list with phases in (0, 1) interval

        """
        super().set_up(x0, data, passband, observer_system_cls=kwargs.get('observer_system_cls'))
        setattr(self, 'discretization', kwargs.pop('discretization'))
        setattr(self, 'interp_treshold', kwargs.pop('interp_treshold'))
        fit_xs = self.generate_sample_phases(kwargs.pop('samples'))
        setattr(self, 'fit_xs', fit_xs)
        self.normalize_data(kind='average')

    @abstractmethod
    def fit(self, *args, **kwargs):
        pass

    def normalize_data(self, kind='global_maximum', top_fraction_to_average=0.1):
        """
        Normalization of the input observational data using different methods. The result is assigned back to the
        respective data attributes `y_data` and `y_err`.

        :param kind: str; normalization method
        :**kind options**:
            * **average** -- each curve is normalized to its average
            * **global_average** -- curves are normalized to their global average
            * **maximum** -- each curve is normalized to its own maximum
            * **global_maximum** -- curves are normalized to their global maximum
        :param top_fraction_to_average: float; top portion of the dataset (in y-axis direction) used in the
                                               normalization process, from (0, 1) interval
        """
        # noinspection PyUnresolvedReferences
        y_data, y_err = normalize_light_curve(self.y_data, self.y_err, kind, top_fraction_to_average)
        setattr(self, 'y_data', y_data)
        setattr(self, 'y_err', y_err)

    def generate_sample_phases(self, samples):
        """
        Sampling photometric phases used to calculate the synthetic observations according to a rule or array of orbital
        phases.

        :param samples: Union[str, List]; `uniform`, `adaptive` or list with phases in (0, 1) interval
        :**kind options**:
            * :'uniform': phase equidistant sampling, use for initial stages of the fitting where a general shape for
                           the solution model is not yet found
            * :'adaptive': equidistant sampling along the synthetic curve, very useful for light curves with narrow
                            eclipses. Use in later stages of fitting where a general shape of the curve is estabilished.
            * :`numpy.array`: array of photometric phases can be set also manually
        :return: Union[numpy.array, ValueError]; photometric phases
        """
        kwargs = parameters.prepare_properties_set(self.initial_vector, self.fitable.keys(), self.constrained,
                                                   self.fixed)
        phases, kwargs = time_layer_resolver(self.x_data_reduced, pop=False, **kwargs)
        if np.shape(phases)[0] < self.interp_treshold:
            return None

        if samples == 'uniform':
            diff = 1.0 / self.interp_treshold
            return np.linspace(0.0 - diff, 1.0 + diff, num=self.interp_treshold + 2)
        elif samples == 'adaptive':
            logger.info('Generating equidistant samples along the light curve using adaptive sampling method')
            return self.adaptive_sampling()
        elif isinstance(samples, (list, np.ndarray)):
            return np.sort(samples)
        else:
            raise ValueError(f'Parameter `samples` has to be either string with values `uniform` or `adaptive` or '
                             f'array of phases in (0, 1) interval')

    def adaptive_sampling(self):
        """
        Function generates sampling equidistantly along the curve defined by the initial vector.

        :return: numpy.array; photometric phases
        """
        n = 3 * settings.MAX_CURVE_DATA_POINTS
        diff = 1.0 / n
        x = np.linspace(0.0 - diff, 1.0 + diff, num=n)

        kwargs = parameters.prepare_properties_set(self.initial_vector, self.fitable.keys(), self.constrained,
                                                   self.fixed)

        kwargs = parameters.extend_json_with_atm_params(
            kwargs,
            atmosphere_model=self.atmosphere_model,
            limb_darkening_coefficients=self.limb_darkening_coefficients
        )

        observer = Observer(passband='bolometric', system=None)
        setattr(observer, 'system_cls', getattr(self.observer, 'system_cls'))
        try:
            synthetic = lc_model.synthetic_binary(x, self.discretization, observer, **kwargs)
            synthetic, _ = normalize_light_curve(synthetic, kind='average')
        except Exception:
            raise RuntimeError('Your initial parameters are invalid and phase sampling could not be generated.')

        curve = np.column_stack((x, synthetic['bolometric']))
        lengths = np.sqrt(np.sum(np.diff(curve, axis=0) ** 2, axis=1))
        crv_lengths = np.cumsum(np.concatenate(([0, ], lengths)))
        segments = np.linspace(0, crv_lengths[-1], num=self.interp_treshold)

        return np.interp(segments, crv_lengths, x)

    def coefficient_of_determination(self, model_parameters, data, discretization, interp_treshold):
        """
        Function returns R^2 for given model parameters and observed data.

        :param model_parameters: Dict; set of model parameters in JSON format
        :param data: Dict[str, elisa.analytics.dataset.base.LCData]; observations in each filter,
        :param discretization: float; discretization factor for the primary component;
        :param interp_treshold: int; Above this total number of datapoints, light curve will be interpolated
                                     using model containing `interp_treshold` equidistant points per epoch
        :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
        """
        self.set_up(x0=parameters.BinaryInitialParameters(**model_parameters), data=data, passband=data.keys(),
                    discretization=discretization, interp_treshold=interp_treshold, samples='uniform',
                    observer_system_cls=BinarySystem)

        r_squared_args = (self.x_data_reduced, self.y_data, self.observer.passband, discretization,
                          self.x_data_reducer, 1.0 / self.interp_treshold, self.interp_treshold,
                          self.observer.system_cls)
        flat_result = parameters.deserialize_result(model_parameters)

        r_dict = {key: value['value'] for key, value in flat_result.items()}
        r_dict = parameters.extend_json_with_atm_params(
            r_dict,
            atmosphere_model=self.atmosphere_model,
            limb_darkening_coefficients=self.limb_darkening_coefficients
        )

        logger.info("Evaluating light curve for calculation of R^2.")
        r_squared_result = lc_r_squared(lc_model.synthetic_binary, *r_squared_args, **r_dict)
        logger.info("Calculation of R^2 finished.")
        return r_squared_result


def lc_r_squared(synthetic, *args, **x):
    """
    Compute R^2 (coefficient of determination) between synthetic and observed light curves.

    :param synthetic: callable; method for crating the synthetic LC observation
                                (eg. elisa.anaytics.models.lc.synthetic_binary)
    :param args: Tuple;
    :**args**:
        * :x_data_reduced: numpy.array; phases in AbstractFit.x_data_reduced slot
        * :y_data: Dict[str, numpy.array]; supplied fluxes (lets say fluxes from observation) normalized to max value
        * :passband: Union[str, List[str]]; list of used photometric filters
        * :discretization: float; discretization factor for the primary component
        * :x_data_reducer: Dict[str, numpy.array]; mask stored in AbstractFit.x_data_reducer slot
        * :diff: float; auxiliary parameter for interpolation defining the spacing between evaluation phases if
                        interpolation is active
        * :interp_treshold: int; a number of observation points above which the synthetic curves will be calculated
                                 using `interp_treshold` equally spaced points that will be subsequently
                                 interpolated to the desired times of observation
        * :cls: BinarySystem;

    :param x: Dict; kwargs containing parameters to compute binary system, keyword arguments of `synthetic` function
    :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
    """
    x_data_reduced, y_data, passband, discretization, x_data_reducer, diff, interp_treshold, cls = args

    x_data_reduced, kwargs = time_layer_resolver(x_data_reduced, pop=False, **x)
    fit_xs = np.linspace(np.min(x_data_reduced) - diff, np.max(x_data_reduced) + diff, num=interp_treshold + 2) \
        if np.shape(x_data_reduced)[0] > interp_treshold else x_data_reduced

    observer = Observer(passband=passband, system=None)
    observer._system_cls = cls
    synthetic = synthetic(fit_xs, discretization, observer, **x)

    if np.shape(fit_xs) != np.shape(x_data_reduced):
        synthetic = {
            fltr: interpolate.interp1d(fit_xs, curve, kind='cubic')(x_data_reduced)
            for fltr, curve in synthetic.items()
        }
    synthetic = {band: synthetic[band][x_data_reducer[band]] for band in synthetic}
    synthetic, _ = normalize_light_curve(synthetic, kind='average')

    return r_squared(synthetic, y_data)


def rv_r_squared(synthetic, *args, **x):
    """
    Compute R^2 (coefficient of determination) between synthetic and observed RV curves.

    :param synthetic: callable; method for crating the synthetic RV observation
                                (eg. elisa.anaytics.models.rv.central_rv_synthetic)
    :param args: Tuple;
    :**args**:
        * :x_data_reduced: numpy.array; phases in AbstractFit.x_data_reduced slot
        * :y_data: Dict[str, numpy.array]; radial velocities for both components
        * :x_data_reducer: Dict[str, numpy.array]; mask stored in AbstractFit.x_data_reducer slot
        * :cls: Union[BinarySystem, RadialVelocitySystem];
    :param x: Dict; kwargs containing parameters to compute radial velocities curve,
                    keyword arguments of `synthetic` function
    :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
    """
    x_data_reduced, y_data, x_data_reducer, cls = args

    observer = Observer(passband='bolometric', system=None)
    observer._system_cls = cls
    synthetic = synthetic(x_data_reduced, observer, **x)
    synthetic = {comp: synthetic[comp][x_data_reducer[comp]] for comp in synthetic}

    return r_squared(synthetic, y_data)


def r_squared(synthetic, observed):
    """
    Returns coefficient of determination between model and observations

    :param synthetic: Dict[str, numpy.array];
    :param observed: Dict[str, numpy.array];
    :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
    """
    variability = np.sum([np.sum(np.power(observed[item] - np.mean(observed[item]), 2)) for item in observed])
    residual = np.sum([np.sum(np.power(synthetic[item] - observed[item], 2)) for item in observed])

    return 1 - (residual / variability)


def extend_observations_to_desired_interval(start_phase, stop_phase, x_data, y_data, y_err):
    """
    Extending the left and the right boundaries of the phase-folded observations to the desired phase interval.

    :param start_phase: float;
    :param stop_phase: float;
    :param x_data: Dict;
    :param y_data: Dict;
    :param y_err: Dict;
    :return: Tuple;
    """
    for item, curve in x_data.items():
        phases_extended = np.concatenate((x_data[item] - 1.0, x_data[item], x_data[item] + 1.0))
        phases_extended_filter = np.logical_and(start_phase < phases_extended, phases_extended < stop_phase)
        x_data[item] = phases_extended[phases_extended_filter]

        y_data[item] = np.tile(y_data[item], 3)[phases_extended_filter]
        if y_err[item] is not None:
            y_err[item] = np.tile(y_err[item], 3)[phases_extended_filter]

    return x_data, y_data, y_err


def check_for_boundary_surface_potentials(result_dict, morphology=None):
    """
    Function checks if surface potential are within errors below critical potentials (which would break BinarySystem
    initialization). If surface potential are below the critical value but still within errors, they are snapped to
    critical potential.

    :param result_dict: Dict; flat dict of fit results
    :param morphology: str; expected morphology
    :return: Dict; corrected flat dict of fit results
    """
    if "primary@surface_potential" not in result_dict.keys() or "secondary@surface_potential" not in result_dict.keys():
        return result_dict

    if result_dict['primary@surface_potential']['value'] == result_dict['secondary@surface_potential']['value']:
        return result_dict

    for component in settings.BINARY_COUNTERPARTS:
        pot = result_dict[f"{component}@surface_potential"]
        if "fixed" not in pot.keys() or "value" not in pot.keys():
            continue

        sigma = pot["value"] - pot["confidence_interval"]["min"] if "confidence_interval" in pot.keys() else 0.001

        synchronicity = result_dict[f"{component}@synchronicity"]["value"] \
            if f"{component}@synchronicity" in result_dict.keys() else 1.0

        mass_ratio = result_dict["system@mass_ratio"]["value"] \
            if "system@mass_ratio" in result_dict.keys() \
            else result_dict["secondary@mass"]["value"]/result_dict["primary@mass"]["value"]

        periastron_distance = 1 - result_dict["system@eccentricity"]["value"]

        l1 = \
            BinarySystem.critical_potential_static(
                component=component,
                components_distance=periastron_distance,
                mass_ratio=mass_ratio,
                synchronicity=synchronicity
            )

        # if resulting potential is too close critical potentials (within errors), it will snap potential to critical
        # to avoid problems
        if 5 * sigma >= l1 - pot["value"] >= 0.0:
            pot["value"] = l1 + 1e-5 * sigma

        # test for over-contact overflow trough L2 point
        l2 = BinarySystem.libration_potentials_static(periastron_distance, mass_ratio)[2]
        if 5 * sigma >= l2 - pot["value"] >= 0.0:
            pot["value"] = l2 - 1e-5 * sigma

    if morphology == 'over-contact':
        result_dict['secondary@surface_potential']['value'] = result_dict['primary@surface_potential']['value']

    return result_dict


def eval_constraint_in_dict(input_dict):
    """
    Evaluates constraints defined supplied in the user given model parameters .

    :param input_dict: Dict; standard JSON format of model parameters
    :return: Dict; same as `input_dict` but with the values added/updated to the constrained parameters according to
                   the values of dependent variable model parameters.
    """
    input_dict1 = parameters.deserialize_result(input_dict)
    result_dict = {key: val for key, val in input_dict1.items() if 'fixed' in val}

    reduced_dict = {key: val['value'] for key, val in result_dict.items()}

    b_parameters = BinaryInitialParameters(**input_dict)
    constraints = b_parameters.get_constrained(jsonify=False)

    constrained_values = parameters.constraints_evaluator(reduced_dict, constraints)
    result_dict.update(
        {
            key: {
                'value': constrained_values[key],
                'constraint': constraints[key].constraint,
                'unit': constraints[key].unit
            } for key, val in constrained_values.items()
        })

    result_dict = parameters.serialize_result(result_dict)
    return result_dict



from ... pulse.container_ops import (
    incorporate_pulsations_to_model,
    generate_harmonics,
)


def build_harmonics(system):
    """
    Adds pre-calculated harmonics for the respective pulsation modes.

    :param system: elisa.single_system.container.SinglePositionContainer;
    """
    if system.star.has_pulsations():
        system.star = generate_harmonics(system.star, com_x=0, phase=system.position.phase, time=system.time)

    return system


def build_perturbations(system):
    """
    Function adds perturbation to the surface mesh due to pulsations.

    :param system: elisa.single_system.container.SinglePositionContainer; instance
    :return: elisa.single_system.container.SinglePositionContainer; instance
    """
    if system.star.has_pulsations():
        args = system.star, 0.0, 1.0
        system.star = incorporate_pulsations_to_model(*args)
    return system



import numpy as np

from ....base.types import FLOAT, INT
from .... pypex.base.conf import PRECISION, ROUND_PRECISION
from .... pypex.poly2d.intersection import sat
from .... pypex.poly2d.point import Point
from .... pypex.utils import det_2d, multiple_determinants


def intersection(p1, p2, p3, p4, in_touch=False, tol=PRECISION, round_tol=ROUND_PRECISION):
    """
    defs::

        x1, y1 = p1 + u * (p2 - p1) = p1 + u * dp1
        x2, y2 = a + v * (b - a) = a + v * dp2
        dp1 = p2 - p1 = (p2_x - p1_x, p2_y - p1_y)
        dp2 = pt4 - pt3 = (pt4_x - pt3_x, pt4_y - pt3_y)

    intersection::

        x1, y1 = x2, y2
        p1 + u * dp1 = a + v * dp2
        in coo
        p1_x + u * dp1_x = pt3_x + v * dp2_x
        p1_y + u * dp1_y = pt3_y + v * dp2_y

        variables::
            u, v

        solution::

            d = (dp1_x * dp2_y) - (dp1_y * dp2_x)
            u = (((p1_y - pt3_y) * dp2_x) - (dp2_y * (p1_x - pt3_x))) / d
            v = (((p1_y - pt3_y) * dp1_x) - (dp1_y * (p1_x - pt3_x))) / d

    :param p1: numpy.array; first point of first segment
    :param p2: numpy.array; second point of first segment
    :param p3: numpy.array; first point of second segment
    :param p4: numpy.array; second point of second segment
    :param round_tol: int; consider two numbers as same if match up to `round_tol` decimal numbers
    :param tol: float; consider number as zero if smaller than 'tol'
    :param in_touch: bool
    :return: tuple

        0: intersection_status::

              False: parallel
              True:  intersection

        1: segment intersection (if segments share common point/s)::

              False:     no intersection
              True:      intersection between defined points or overlap
              numpy.nan: uknown

        2: intersection Point

        3: distance if parallel
        
        4: string representation/description
    """

    p1, p2, p3, p4 = np.array(p1), np.array(p2), np.array(p3), np.array(p4)
    # first line
    dp1 = p2 - p1
    # second line
    dp2 = p4 - p3
    # determinant
    matrix = np.array([dp1, dp2])
    d = det_2d(matrix)

    # test if d < 1e-10
    # testing on zero, but precission should cause p3 problem
    if np.abs(d) < tol:
        # test distance between lines
        # if general form is known (ax + by + c1 = 0 and ax + by + c2 = 0),
        # d = abs(c1 - c2) / sqrt(a**2 + b**2)
        # parametric equation in general:
        #   x, y = [p1_x, p1_y] + u * [T_x, T_y], where T is tangential vector defined as p2 - p1
        # N = (a, b) represent normal vector of line; `p3`, `p4` (method parametres) from general equation of line
        # N = [-Ty, Tx], can be obtained
        # general equation:
        #   -Ty * x + Tx * y + c = 0, then
        # c = Ty * p1_x - Tx * p1_y
        # finaly, general equation:
        #   -Ty * x + Tx * y + (Ty * p1_x - Tx * p1_y) = 0
        #
        #
        # a1, b1, c1 = -dp1_y, dp1_x, (dp1_y * pt1_x) - (dp1_x * pt1_y)
        # a2, b2, c2 = -dp2_y, dp2_x, (dp2_y * pt3_x) - (dp2_x * pt3_y)

        a1, b1, c1 = -dp1[1], dp1[0], det_2d(np.array([p1, dp1]))

        # second line has to be definable with same tangential and normal vector as first line
        # since ax + by + c = 0 and in our case [x, y] = p3 or p4 for second equation, then for c2
        # we have c2 = - (a1 * p3[0] + b1 * p3[1])
        c2 = - (a1 * p3[0] + b1 * p3[1])
        d = abs(c2 - c1) / (np.sqrt(a1 ** 2 + b1 ** 2))

        intersects, msg = (True, "OVERLAP") if abs(d) < tol else (False, "PARALLEL")
        int_in_segment = False if msg in ["PARALLEL"] \
            else sat.intersects(np.array([p1, p2]), np.array([p3, p4]), in_touch, round_tol)
        return intersects, int_in_segment, np.nan, d, msg

    # +0 because of negative zero (-0.0 is incorrect) formatting on output
    u = (det_2d([dp2, p1 - p3]) / d) + 0.
    v = (det_2d([dp1, p1 - p3]) / d) + 0.

    eval_method = np.less_equal if in_touch else np.less
    int_x, int_y = p1[0] + (u * dp1[0]), p1[1] + (u * dp1[1])
    int_segment = True if np.logical_and(eval_method(0.0, u), eval_method(u, 1.0)) and \
        np.logical_and(eval_method(0.0, v), eval_method(v, 1.0)) else False
    return True, int_segment, Point(int_x, int_y), np.nan, "INTERSECT"


def intersections(poly1, poly2, in_touch=False, tol=PRECISION, round_tol=ROUND_PRECISION):
    """
    Vectorised implementaion of lines intersection function. Compute intersections of all combination of supplied
    arrays of points which define convex polygon

    :param poly1: numpy.array; clokwise ordered numpy array of points
    :param poly2: numpy.array; clokwise ordered numpy array of points
    :param in_touch: bool; consider touch in one point as intersection
    :param tol: consider all numbers as zero when abs(number) < tol
    :param round_tol: consider two numbers as same if match up to round_tol deciaml numbers
    :return: tuple;
    """
    m1, _ = poly1.shape
    m2, _ = poly2.shape
    n1, n2 = 2, 2

    # mask to origin
    idx_mask = _index_map(m1, m2)

    intersection_status, intersection_segment = np.zeros(m1 * m2, dtype=bool), np.zeros(m1 * m2, dtype=bool)
    intr_ptx = np.full_like(np.empty((m1 * m2, 2), dtype=FLOAT), np.nan)
    distance = np.full_like(np.empty(m1 * m2, dtype=FLOAT), np.nan)

    msg = np.chararray(m1 * m2, itemsize=9)

    poly1_edges = polygon_hull_to_edges(poly1)
    poly2_edges = polygon_hull_to_edges(poly2)

    dif_poly1 = poly1_edges[:, 1, :] - poly1_edges[:, 0, :]
    dif_poly2 = poly2_edges[:, 1, :] - poly2_edges[:, 0, :]

    # make all possible determinants matrix for all combination of lines (needs for equation solver 1/D)
    corr_dpoly1 = np.repeat(dif_poly1, m2, axis=0)
    corr_dpoly2 = np.tile(dif_poly2, (m1, 1))

    det_matrix = np.empty((m1 * m2, n1, n2))
    det_matrix[:, 0, :] = corr_dpoly1
    det_matrix[:, 1, :] = corr_dpoly2

    determinants = multiple_determinants(det_matrix)
    non_intersections = np.array(np.abs(determinants) < tol)

    if non_intersections.any():
        problem_poly1 = np.repeat(poly1, m2, axis=0)[non_intersections]
        problem_dif_poly1 = np.repeat(dif_poly1, m2, axis=0)[non_intersections]
        a1, b1 = -problem_dif_poly1[:, 1], problem_dif_poly1[:, 0]

        face_dface = np.empty((problem_poly1.shape[0], 2, problem_poly1.shape[1]), dtype=problem_poly1.dtype)
        face_dface[:, 0, :] = problem_poly1
        face_dface[:, 1, :] = problem_dif_poly1
        c1 = multiple_determinants(face_dface)
        problem_poly2_edges = np.tile(poly2_edges, (poly1.shape[0], 1, 1))[non_intersections]
        c2 = -(a1 * problem_poly2_edges[:, 1, 0] + b1 * problem_poly2_edges[:, 1, 1])
        dist = np.abs(c2 - c1) / (np.sqrt(np.power(a1, 2) + np.sqrt(np.power(b1, 2))))
        # fill output
        distance[non_intersections] = dist
        msg[non_intersections] = str('PARALLEL')
        overlaps = non_intersections.copy()

        overlaps[non_intersections] = np.abs(dist) < tol
        intersection_status[overlaps] = True

        if np.any(overlaps):
            # assume that in real life, there will neglible amount of parallel lines with zero distance (overlap lines)
            # so we can use for loop without any significant loose of performance
            poly1_comb_overlap = np.repeat(poly1_edges, m2, axis=0)[overlaps]
            poly2_comb_overlap = np.tile(poly2_edges, (m1, 1, 1))[overlaps]

            intersection_segment[overlaps] = np.array([sat.intersects(a, b, in_touch=in_touch, round_tol=round_tol)
                                                       for a, b in zip(poly1_comb_overlap, poly2_comb_overlap)])
            msg[overlaps] = 'OVERLAP'

    ints = ~non_intersections
    ok_dif_poly1, ok_dif_poly2 = corr_dpoly1[ints], corr_dpoly2[ints]
    ok_poly1_edges, ok_poly2_edges = np.repeat(poly1_edges, m2, axis=0)[ints], np.tile(poly2_edges, (m1, 1, 1))[ints]

    p1_p3 = ok_poly1_edges[:, 0, :] - ok_poly2_edges[:, 0, :]

    dp2_p1_p3matrix = _dpx_p1_p3matrix(p1_p3, ok_dif_poly1, ok_dif_poly2)
    dp1_p1_p3matrix = _dpx_p1_p3matrix(p1_p3, ok_dif_poly1, ok_dif_poly1)

    d = determinants[ints]
    u = (multiple_determinants(dp2_p1_p3matrix) / d) + 0.0
    v = (multiple_determinants(dp1_p1_p3matrix) / d) + 0.0

    eval_method = np.less_equal if in_touch else np.less
    intersect_in = ok_poly1_edges[:, 0, :] + (u[:, np.newaxis] * ok_dif_poly1)

    u_in_range = np.logical_and(eval_method(0.0, u), eval_method(u, 1.0))
    v_in_range = np.logical_and(eval_method(0.0, v), eval_method(v, 1.0))
    segments_intersection_status = np.logical_and(u_in_range, v_in_range)

    # fill output
    intersection_status[ints] = True
    intersection_segment[ints] = segments_intersection_status
    msg[ints] = 'INTERSECT'
    intr_ptx[ints] = intersect_in

    return intersection_status, intersection_segment, intr_ptx, distance, msg, idx_mask


def _index_map(m1, m2):
    x = np.empty((m1, 2), dtype=INT)
    x[:, 0] = np.arange(m1)
    x[:, 1] = np.roll(x[:, 0], axis=0, shift=-1)

    y = np.empty((m2, 2))
    y[:, 0] = np.arange(m2)
    y[:, 1] = np.roll(y[:, 0], axis=0, shift=-1)

    idx_map = np.empty((m1 * m2, 4))
    idx_map[:, :2] = np.repeat(x, m2, axis=0)
    idx_map[:, 2:] = np.tile(y, (m1, 1))
    return idx_map


def polygon_hull_to_edges(hull: np.array):
    edges = np.zeros((hull.shape[0], 2, 2))
    edges[:, 0, :] = hull
    edges[:, 1, :] = np.roll(hull, axis=0, shift=-1)
    return edges


def _dpx_p1_p3matrix(p1_p3, dpoly1, dpoly2):
    dpx_p1_p3matrix = np.empty((dpoly1.shape[0], 2, p1_p3.shape[1]), dtype=p1_p3.dtype)
    dpx_p1_p3matrix[:, 0, :] = dpoly2
    dpx_p1_p3matrix[:, 1, :] = p1_p3
    return dpx_p1_p3matrix



import numpy as np
from .. import umpy as up


def static_potential_primary_fn(radius, *args):
    """
    Potential function which defines surface of primary component of binary system in spherical coordinates.
    It is exposed for multiprocessing to avoid pickling of classes, loggers, etc.

    :param radius: float;
    :param args: Tuple; (mass_ratio, surface_potential, b, c, d, e)

    (b, c, d, e) are precalculated values to decrease runtime::

        (b, c, d, e) such that: Psi1 = 1/r + a/sqrt(b+r^2+c*r) - d*r + e*x^2

    :return: float;
    """
    mass_ratio, surface_potential, b, c, d, e = args
    radius2 = up.power(radius, 2)
    a = 1 / radius + mass_ratio / up.sqrt(b + radius2 - c * radius) - d * radius + e * radius2
    return a - surface_potential


def static_potential_secondary_fn(radius, *args):
    """
    Potential function which defines surface of primary component of binary system in spherical coordinates.
    It is exposed for multiprocessing to avoid pickling of classes, loggers, etc.

    :param radius: float;
    :param args: Tuple; (mass_ratio, surface_potential, b, c, d, e, f)

    (b, c, d, e, f) are precalculated values to decrease runtime::

        (b, c, d, e, f) such that: Psi2 = q/r + 1/sqrt(b+r^2-Cr) - d*r + e*x^2 + f

    :return: float;
    """
    mass_ratio, surface_potential, b, c, d, e, f = args
    radius2 = up.power(radius, 2)
    a = mass_ratio / radius + 1. / up.sqrt(b + radius2 - c * radius) - d * radius + e * radius2 + f
    return a - surface_potential


def static_potential_primary_cylindrical_fn(radius, *args):
    """
    Potential function which defines surface of primary component of binary system in cylindrical coordinates.
    Useful for W UMa systems.

    :param radius: float;
    :param args: Tuple; (mass_ratio, surface_potential, a, b, c, d, e, f)

    (a, b, c, d, e, f) are precalculated values to decrease runtime::

        (a, b, c, d, e, f) such that: Psi1 = 1/sqrt(a+r^2) + q/sqrt(b + r^2) - c + d*(e+f*r^2)

    :return: float;
    """
    mass_ratio, surface_potential, a, b, c, d, e, f = args
    radius2 = up.power(radius, 2)
    return 1 / up.sqrt(a + radius2) + mass_ratio / up.sqrt(b + radius2) - c + d * (e + f * radius2) - surface_potential


def static_potential_secondary_cylindrical_fn(radius, *args):
    """
    Potential function which defines surface of primary component of binary system in cylindrical coordinates.
    Useful for W UMa systems.

    :param radius: float;
    :param args: Tuple; (mass_ratio, surface_potential, a, b, c, d, e, f)

    (a, b, c, d, e, f) are precalculated values to decrease runtime::

        (a, b, c, d, e, f) such that: Psi1 = 1/sqrt(a+r^2) + q/sqrt(b + r^2) - c + d*(e+f*r^2)

    :return: float;
    """
    mass_ratio, surface_potential, a, b, c, d, e, f = args
    radius2 = up.power(radius, 2)
    return mass_ratio / up.sqrt(a + radius2) + 1. / up.sqrt(b + radius2) + c * (d + e * radius2) + f - surface_potential


def potential_value_primary(radius, *args):
    """
    Calculates modified Kopal's potential from point of view of primary component.

    :param radius: (numpy.)float; spherical variable
    :param args: Tuple; (mass_ratio, B, C, D, E) such that: Psi1 = 1/r + q/sqrt(B+r^2+Cr) - D*r + E*r^2
    :return: (numpy.)float;
    """
    mass_ratio, b, c, d, e = args
    radius2 = up.power(radius, 2)
    return 1 / radius + mass_ratio / up.sqrt(b + radius2 - c * radius) - d * radius + e * radius2


def potential_value_secondary(radius, *args):
    """
    Calculates modified Kopal's potential from point of view of secondary component.
    :param radius: up.float; spherical variable
    :param args: Tuple; (mass_ratio, b, c, d, e, f) such that: Psi2 = q/r + 1/sqrt(b+r^2-Cr) - d*r + e*r^2 + f
    :return: float;
    """
    mass_ratio, b, c, d, e, f = args
    radius2 = up.power(radius, 2)
    return mass_ratio / radius + 1. / up.sqrt(b + radius2 - c * radius) - d * radius + e * radius2 + f


def potential_primary_fn(radius, *args):
    """
    Implicit potential function from perspective of primary component.

    :param radius: float; spherical variable
    :param args: Tuple; pre calculated values for potential function and desired value of potential
    :return: float;
    """
    return potential_value_primary(radius, *args[0]) - args[1]


def potential_secondary_fn(radius, *args):
    """
    Implicit potential function from perspective of secondary component.

    :param radius: float; spherical variable
    :param args: pre calculated values for potential function and desired value of potential
    :return: float;
    """
    # return self.potential_value_secondary(radius, *args) - self.secondary.surface_potential
    return potential_value_secondary(radius, *args[0]) - args[1]


def potential_value_primary_cylindrical(radius, *args):
    """
    Calculates modified Kopal's potential from point of view of primary component
    in cylindrical coordinates r_n, phi_n, z_n, where z_n = x and heads along z axis.

    This function is intended for generation of ``necks``
    of W UMa systems, therefore components distance = 1 an synchronicity = 1 is assumed.

    :param radius: float;
    :param args: Tuple; (a, b, c, d, e) such that: Psi1 = 1/sqrt(a+r^2) + q/sqrt(b + r^2) - c + d*(a+e*r^2)
    :return: float;
    """
    mass_ratio, a, b, c, d, e = args

    radius2 = up.power(radius, 2)
    return 1 / up.sqrt(a + radius2) + mass_ratio / up.sqrt(b + radius2) - c + d * (a + e * radius2)


def potential_value_secondary_cylindrical(radius, *args):
    """
    Calculates modified kopal potential from point of view of secondary
    component in cylindrical coordinates r_n, phi_n, z_n, where z_n = x and heads along z axis.

    :param radius: float;
    :param args: Tuple; (a, b, c, d, e, f) such that: Psi2 = q/sqrt(a+r^2) + 1/sqrt(b + r^2) - c + d*(a+e*r^2)
    :return: float;
    """
    mass_ratio, a, b, c, d, e, f = args

    radius2 = up.power(radius, 2)
    return mass_ratio / up.sqrt(a + radius2) + 1. / up.sqrt(b + radius2) - c + d * (e * radius2 + a) + f


def potential_primary_cylindrical_fn(radius, *args):
    """
    Implicit potential function from perspective of primary component given in cylindrical coordinates.

    :param radius: float;
    :param args: Tuple; pre calculated values for potential function and desired value of potential
    :return: float;
    """
    return potential_value_primary_cylindrical(radius, *args[0]) - args[1]


def potential_secondary_cylindrical_fn(radius, *args):
    """
    Implicit potential function from perspective of secondary component given in cylindrical coordinates.

    :param radius: float;
    :param args: Tuple: pre calculated values for potential function and desired value of potential
    :return: float;
    """
    return potential_value_secondary_cylindrical(radius, *args[0]) - args[1]


def radial_primary_potential_derivative(radius, *args):
    """
    Function calculate radial derivative of primary potential function in spherical coordinates.

    :param radius: radius of given point(s) in spherical coordinates
    :type radius: Union[float, numpy.array];
    :param args: b, c, d, e - such that: dPsi1/dr = -1/r^2 + 0.5*q*(c-2r)/(b-cr+r^2)^(3/2) - d +2er
    :type args: Tuple;
    :return: Union[float, numpy.array];
    """
    # auxiliary values pre-calculated in pre_calculate_for_potential_value_primary()
    mass_ratio, b, c, d, e = args
    radius2 = up.power(radius, 2)

    return - 1 / radius2 + 0.5 * mass_ratio * (c - 2 * radius) / up.power(b - c * radius + radius2, 1.5) \
           - d + 2 * e * radius


def radial_secondary_potential_derivative(radius, *args):
    """
    Function calculate radial derivative of secondary potential function in spherical coordinates

    :param radius: radius of given point(s) in spherical coordinates
    :type radius: Union[float, numpy.array];
    :param args: b, c, d, e, f - such that: dPsi2/dr = -q/r^2 + (0.5c - x)/(b - cx + r^2)^(3/2) - d + 2er
    :type args: Tuple;
    :return: float;
    """
    # auxiliary values pre-calculated in pre_calculate_for_potential_value_primary()
    mass_ratio, b, c, d, e, f = args
    radius2 = up.power(radius, 2)

    return - mass_ratio / radius2 + (0.5 * c - radius) / up.power(b - c * radius + radius2, 1.5) \
           - d + 2 * e * radius


def pre_calculate_for_potential_value_primary(*args, return_as_tuple=False):
    """
    Function calculates auxiliary values for calculation of primary component potential,
    and therefore they don't need to be wastefully recalculated every iteration in solver.

    :param return_as_tuple: return coefficients as a tuple of numpy vectors instead of numpy matrix
    :type return_as_tuple: bool;
    :param args: (component distance, azimut angle (0, 2pi), latitude angle (0, pi)
    :return: Tuple; (b, c, d, e) such that: Psi1 = 1/r + q/sqrt(b+r^2-c*r) - d*r + e*r^2
    """
    # synchronicity, mass_ratio, distance between components, azimuth angle, latitude angle (0,180)
    synchronicity, mass_ratio, distance, phi, theta = args

    sin_theta = up.sin(theta)
    cs = up.cos(phi) * sin_theta

    b = up.power(distance, 2)
    c = 2 * distance * cs
    d = (mass_ratio * cs) / b
    e = 0.5 * up.power(synchronicity, 2) * (1 + mass_ratio) * up.power(sin_theta, 2)

    if np.isscalar(phi):
        return b, c, d, e
    else:
        bb = b * np.ones(np.shape(phi))
        return (bb, c, d, e) if return_as_tuple else np.column_stack((bb, c, d, e))


def pre_calculate_for_potential_value_secondary(*args, return_as_tuple=False):
    """
    Function calculates auxiliary values for calculation of secondary component potential,
    and therefore they don't need to be wastefully recalculated every iteration in solver.

    :param return_as_tuple: return coefficients as a Tuple of numpy vectors instead of numpy matrix
    :type return_as_tuple: bool
    :param args: (component distance, azimut angle (0, 2pi), latitude angle (0, pi)
    :return: Tuple: (b, c, d, e, f) such that: Psi2 = q/r + 1/sqrt(b+r^2+Cr) - d*r + e*r^2 + f
    """
    # synchronicity, mass_ratio, distance between components, azimuth angle, latitude angle (0,180)
    synchronicity, mass_ratio, distance, phi, theta = args

    sin_theta = up.sin(theta)
    cs = up.cos(phi) * sin_theta

    b = up.power(distance, 2)
    c = 2 * distance * cs
    d = cs / b
    e = 0.5 * up.power(synchronicity, 2) * (1 + mass_ratio) * up.power(sin_theta, 2)
    f = 0.5 - 0.5 * mass_ratio

    if np.isscalar(phi):
        return b, c, d, e, f
    else:
        bb = b * np.ones(np.shape(phi))
        ff = f * np.ones(np.shape(phi))
        return (bb, c, d, e, ff) if return_as_tuple else np.column_stack((bb, c, d, e, ff))


def radial_primary_potential_derivative_cylindrical(radius, *args):
    """
    Function calculate radial derivative of primary potential function in cylindrical coordinates.

    :param radius: radius of given point(s) in cylindrical coordinates
    :type radius: Union[float, numpy.array]
    :param args: a, b, c, d, e such that: dPsi1/dr = - r/(a+r^2)^(3/2) - rq/(b+r^2)^(3/2) + 2der
    :type args: Tuple;
    :return: float;
    """
    mass_ratio, a, b, c, d, e = args

    radius2 = up.power(radius, 2)
    return + 0 - radius / up.power(a + radius2, 1.5) - radius * mass_ratio / up.power(b + radius2, 1.5) \
           + 2 * d * e * radius


def radial_secondary_potential_derivative_cylindrical(radius, *args):
    """
    Function calculate radial derivative of secondary potential function in cylindrical coordinates.

    :param radius: radius of given point(s) in cylindrical coordinates
    :type radius: Union[float, numpy.array]
    :param args: a, b, c, d, e, f such that: dPsi2/dr = - qr/(a+r^2)^(3/2) - r/(b+r^2)^(3/2) + 2cer
    :type args: Tuple;
    :return: float;
    """
    mass_ratio, a, b, c, d, e, f = args

    radius2 = up.power(radius, 2)
    return + 0 - radius * mass_ratio / up.power(a + radius2, 1.5) - radius / up.power(b + radius2, 1.5) \
           + 2 * d * e * radius


def pre_calculate_for_potential_value_primary_cylindrical(*args, return_as_tuple=False):
    """
    Function calculates auxiliary values for calculation of primary component potential
    in cylindrical symmetry. Therefore they don't need to be wastefully recalculated every iteration in solver.

    :param return_as_tuple: return coefficients as a tuple of numpy vectors instead of numpy matrix
    :type return_as_tuple: bool
    :param args: (azimut angle (0, 2pi), z_n (cylindrical, identical with cartesian x)), components distance
    :return: Tuple; (a, b, c, d, e) such that: Psi1 = 1/sqrt(a+r^2) + q/sqrt(b + r^2) - c + d*(a+e*r^2)
    """
    synchronicity, mass_ratio, phi, z, distance = args

    a = up.power(z, 2)
    b = up.power(distance - z, 2)
    c = mass_ratio * z / up.power(distance, 2)
    d = 0.5 * up.power(synchronicity, 2) * (1 + mass_ratio)
    e = up.power(up.sin(phi), 2)

    if np.isscalar(phi):
        return a, b, c, d, e
    else:
        dd = d * np.ones(np.shape(phi))
        return (a, b, c, dd, e) if return_as_tuple else np.column_stack((a, b, c, dd, e))


def pre_calculate_for_potential_value_secondary_cylindrical(*args, return_as_tuple=False):
    """
    Function calculates auxiliary values for calculation of secondary
    component potential in cylindrical symmetry, and therefore they don't need
    to be wastefully recalculated every iteration in solver.

    :param return_as_tuple: return coefficients as a tuple of numpy vectors instead of numpy matrix
    :type return_as_tuple: bool;
    :param args: (azimut angle (0, 2pi), z_n (cylindrical, identical with cartesian x)), components distance
    :return: Tuple; (a, b, c, d, e, f) such that: Psi2 = q/sqrt(a+r^2) + 1/sqrt(b + r^2) - c + d*(a+e*r^2)
    """
    synchronicity, mass_ratio, phi, z, distance = args

    a = up.power(z, 2)
    b = up.power(distance - z, 2)
    c = z / up.power(distance, 2)
    d = 0.5 * up.power(synchronicity, 2) * (1 + mass_ratio)
    e = up.power(up.sin(phi), 2)
    f = 0.5 * (1 - mass_ratio)

    if np.isscalar(phi):
        return a, b, c, d, e, f
    else:
        dd = d * np.ones(np.shape(phi))
        ff = f * np.ones(np.shape(phi))
        return (a, b, c, dd, e, ff) if return_as_tuple else np.column_stack((a, b, c, dd, e, ff))


def primary_potential_derivative_x(x, *args):
    """
    Derivative of potential function perspective of primary component along the x axis.

    :param x: (numpy.)float;
    :param args: Tuple (float, float, float); (synchronicity of primary component, mass ratio, components distance)
    :return: (numpy.)float;
    """
    synchronicity, mass_ratio, d = args
    r_sqr, rw_sqr = x ** 2, (d - x) ** 2
    return - (x / r_sqr ** (3.0 / 2.0)) + ((mass_ratio * (d - x)) / rw_sqr ** (
            3.0 / 2.0)) + synchronicity ** 2 * (mass_ratio + 1) * x - mass_ratio / d ** 2


def secondary_potential_derivative_x(x, *args):
    """
    Derivative of potential function perspective of secondary component along the x axis.

    :param x: (numpy.)float;
    :param args: Tuple (float, float, float); (synchronicity of secondary component, mass ratio, components distance)
    :return: (numpy.)float;
    """
    synchronicity, mass_ratio, d = args
    r_sqr, rw_sqr = x ** 2, (d - x) ** 2
    return - (x / r_sqr ** (3.0 / 2.0)) + ((mass_ratio * (d - x)) / rw_sqr ** (
            3.0 / 2.0)) - synchronicity ** 2 * (mass_ratio + 1) * (d - x) + (1.0 / d ** 2)



import numpy as np
from copy import copy

from . import utils as crv_utils
from .. surface.mesh import add_spots_to_mesh
from .. import (
    utils as bsutils,
    dynamic
)
from .. container import OrbitalPositionContainer
from .. surface.coverage import compute_surface_coverage
from .. orbit.container import OrbitalSupplements
from ... import utils, const
from ... import settings
from ... base.types import INT


def produce_circ_sync_curves_mp(*args):
    """
    Curve generator function for circular synchronous systems.

    :param args: Tuple;

    ::

        Tuple[
                binary: elisa.binary_system.BinarySystem,
                initial_system: elisa.binary_system.container.OrbitalPositionContainer, system container with built
                geometry
                phase_batch: numpy.array; phases at which to calculate curves,
                normal_radiance: Dict; {component: numpy.array; normal radiances for each surface element},
                ld_cfs: Dict;
                crv_labels: List;
                curves_fn: function to calculate curve points at given orbital positions,
                kwargs: Dict,
            ]

    :return: Dict;
    """
    # todo: too much replicated code from produce_circ_pulsating_curves_mp
    binary, initial_system, phase_batch, crv_labels, curves_fn, kwargs = args

    position_method = kwargs.pop("position_method")
    orbital_motion = position_method(input_argument=phase_batch, return_nparray=False, calculate_from='phase')
    # is in eclipse test eval
    ecl_boundaries = dynamic.get_eclipse_boundaries(binary, 1.0)
    azimuths = [position.azimuth for position in orbital_motion]
    in_eclipse = dynamic.in_eclipse_test(azimuths, ecl_boundaries)

    curves = {key: np.zeros(phase_batch.shape) for key in crv_labels}

    for pos_idx, position in enumerate(orbital_motion):
        on_pos = bsutils.move_sys_onpos(initial_system, position, on_copy=True)

        compute_surface_coverage(on_pos, binary.semi_major_axis, in_eclipse=in_eclipse[pos_idx],
                                 return_values=False, write_to_containers=True)

        curves = curves_fn(curves, pos_idx, crv_labels, on_pos)

    return curves


def produce_circ_spotty_async_curves_mp(*args):
    """
    Curve generator function for circular asynchronous spotty systems.

    :param args: Tuple;

    ::

        Tuple[
                binary: elisa.binary_system.BinarySystem,
                initial_system: elisa.binary_system.container.OrbitalPositionContainer, system container with built
                geometry: surface points of a clean system
                phase_batch: numpy.array; phases at which to calculate curves,
                ecl_boundaries: boundaries for both eclipses
                crv_labels: List;
                curves_fn: function to calculate curve points at given orbital positions,
                kwargs: Dict,
            ]

    :return: Dict;
    """
    binary, initial_system, motion_batch, base_points, ecl_boundaries, crv_labels, curve_fn, kwargs = args

    # pre-calculate the longitudes of each spot for each phase
    phases = np.array([val.phase for val in motion_batch])
    in_eclipse = dynamic.in_eclipse_test([position.azimuth for position in motion_batch], ecl_boundaries)
    spots_longitudes = dynamic.calculate_spot_longitudes(binary, phases, component="all", correct_libration=False)
    pulsation_tests = {'primary': binary.primary.has_pulsations(), 'secondary': binary.secondary.has_pulsations()}
    primary_reducer, secondary_reducer = \
        dynamic.resolve_spots_geometry_update(spots_longitudes, len(phases), pulsation_tests)
    combined_reducer = primary_reducer & secondary_reducer

    # calculating lc with spots gradually shifting their positions in each phase
    curves = {key: np.empty(len(motion_batch)) for key in crv_labels}
    normal_radiance, ld_cfs = None, None
    for pos_idx, orbital_position in enumerate(motion_batch):
        initial_system.set_on_position_params(position=orbital_position)
        initial_system.time = initial_system.set_time()
        # setup component necessary to build/rebuild

        # require_build = "all"
        require_build = "all" if combined_reducer[pos_idx] \
            else "primary" if primary_reducer[pos_idx] \
            else "secondary" if secondary_reducer[pos_idx] \
            else None

        # use clear system surface points as a starting place to save a time
        # if reducers for related component is set to False, previous build will be used

        if primary_reducer[pos_idx]:
            initial_system.primary.points = copy(base_points['primary'])
        if secondary_reducer[pos_idx]:
            initial_system.secondary.points = copy(base_points['secondary'])

        # assigning new longitudes for each spot
        dynamic.assign_spot_longitudes(initial_system, spots_longitudes, index=pos_idx, component=require_build)

        # build the spots points
        add_spots_to_mesh(initial_system, orbital_position.distance, component=require_build)
        # build the rest of the surface based on preset surface points
        _build_args = dict(components_distance=orbital_position.distance, component=require_build)
        initial_system.build_faces_and_kinematic_quantities(**_build_args)
        initial_system.build_temperature_distribution(components_distance=orbital_position.distance, component='all')

        if initial_system.has_pulsations():
            on_pos = initial_system.copy()
            on_pos.flat_it()
            on_pos.build_pulsations(components_distance=orbital_position.distance, component='all')
            on_copy, sys_to_rotate = False, on_pos
        else:
            on_copy, sys_to_rotate = True, initial_system

        on_pos = bsutils.move_sys_onpos(sys_to_rotate, orbital_position, on_copy=on_copy)

        # if None of components has to be rebuilt, use previously computed radiances and limbdarkening when available
        require_build_test = require_build is not None
        on_pos, normal_radiance, ld_cfs = \
            crv_utils.update_surface_params(require_build_test, on_pos, normal_radiance, ld_cfs, **kwargs)

        _kwargs = dict(in_eclipse=in_eclipse[pos_idx], return_values=False, write_to_containers=True)
        compute_surface_coverage(on_pos, binary.semi_major_axis, **_kwargs)

        curves = curve_fn(curves, pos_idx, crv_labels, on_pos)

    return curves


def produce_circ_pulsating_curves_mp(*args):
    """
    Curve generator function for circular pulsating systems.

    :param args: Tuple;

    ::

        Tuple[
                binary: elisa.binary_system.BinarySystem,
                initial_system: elisa.binary_system.container.OrbitalPositionContainer, system container with built
                phase_batch: numpy.array; phases at which to calculate curves,
                crv_labels: List;
                curves_fn: function to calculate curve points at given orbital positions,
                kwargs: Dict,
            ]

    :return:
    """
    # todo: too much replicated code from produce_circ_sync_curves_mp
    binary, initial_system, phase_batch, crv_labels, curves_fn, kwargs = args

    position_method = kwargs.pop("position_method")
    orbital_motion = position_method(input_argument=phase_batch, return_nparray=False, calculate_from='phase')
    # is in eclipse test eval
    ecl_boundaries = dynamic.get_eclipse_boundaries(binary, 1.0)
    azimuths = [position.azimuth for position in orbital_motion]
    in_eclipse = dynamic.in_eclipse_test(azimuths, ecl_boundaries)

    curves = {key: np.zeros(phase_batch.shape) for key in crv_labels}

    for pos_idx, position in enumerate(orbital_motion):
        on_pos = initial_system.copy()
        on_pos.set_on_position_params(position)
        on_pos.set_time()

        on_pos.build_pulsations(components_distance=position.distance)
        crv_utils.prep_surface_params(on_pos, return_values=False, write_to_containers=True, **kwargs)
        on_pos = bsutils.move_sys_onpos(on_pos, position, on_copy=False, recalculate_velocities=False)

        compute_surface_coverage(on_pos, binary.semi_major_axis, in_eclipse=in_eclipse[pos_idx],
                                 return_values=False, write_to_containers=True)

        curves = curves_fn(curves, pos_idx, crv_labels, on_pos)

    return curves


def integrate_eccentric_curve_exactly(*args):
    """
    Curve generator function for exact integration of eccentric orbits.

    :param args: Tuple;

    ::

        Tuple[
                binary: elisa.binary_system.BinarySystem,
                potentials: Dict; corrected surface potentials
                phase_batch: numpy.array; phases at which to calculate curves,
                spots_longitudes: longitudes of each spots for each orbital position
                curves_fn: function to calculate curve points at given orbital positions,
                kwargs: Dict,
            ]

    :return: Dict;
    """
    binary, potentials, motion_batch, spots_longitudes, crv_labels, curve_fn, kwargs = args
    curves = {key: np.empty(len(motion_batch)) for key in crv_labels}
    for run_idx, position in enumerate(motion_batch):
        pos_idx = int(position.idx)
        from_this = dict(binary_system=binary, position=position)
        on_pos = OrbitalPositionContainer.from_binary_system(**from_this)
        dynamic.assign_spot_longitudes(on_pos, spots_longitudes, index=pos_idx, component="all")
        on_pos.set_on_position_params(position, potentials["primary"][pos_idx], potentials["secondary"][pos_idx])
        on_pos.build(components_distance=position.distance)
        on_pos = bsutils.move_sys_onpos(on_pos, position, on_copy=False)

        crv_utils.prep_surface_params(on_pos, return_values=False, write_to_containers=True, **kwargs)
        # TODO: properly calculate in_eclipse parameter
        _kwargs = dict(in_eclipse=True, return_values=False, write_to_containers=True)
        compute_surface_coverage(on_pos, binary.semi_major_axis, **_kwargs)

        curves = curve_fn(curves, run_idx, crv_labels, on_pos)
    return curves


# managing approximations in eccentric orbits ##########################################################################

def _update_surface_in_ecc_orbits(system, orbital_position, new_geometry_test):
    """
    Function decides how to update surface properties with respect to the degree of change
    in surface geometry given by new_geometry test.
    If true, only points and normals are recalculated, otherwise surface is calculated from scratch.

    :param system: elisa.binary_system.container.OrbitalPositionContainer
    :param orbital_position:  OrbitalPosition list
    :param new_geometry_test: bool; test that will decide, how the following phase will be calculated
    :return: elisa.binary_system.system.BinarySystem; instance with updated geometry
    """
    if new_geometry_test:
        system.build(components_distance=orbital_position.distance)
    else:
        system.rebuild_symmetric_detached_mesh(component="all", components_distance=orbital_position.distance)
        system.build_velocities(components_distance=orbital_position.distance, component="all")
        system.build_faces_orientation(component="all", components_distance=orbital_position.distance)
        system.correct_mesh(component="all", components_distance=orbital_position.distance)
        system.build_surface_areas(component="all")

    return system


def _update_ldc_and_radiance_on_orb_pair(
        new_geometry_test,
        base_container,
        mirror_container,
        old_normal_radiance,
        old_ld_cfs,
        **kwargs
):
    """
    Function recalculates or assigns old values to normal radiances or limb darkening coefficients.

    :param new_geometry_test: bool; if True, parameters will be recalculated according
                                    to new geometry, otherwise they will be copied
    :param base_container: elisa.binary_system.container.OrbitalPositionContainer;
    :param mirror_container: elisa.binary_system.container.OrbitalPositionContainer;
    :param old_normal_radiance: Dict; normal radiances to be copied if `new_geometry_test` is False
    :param old_ld_cfs: Dict; normal radiances to be copied if `new_geometry_test` is False
    :param kwargs: kwargs;
    :return: Tuple;
    """
    if new_geometry_test:
        normal_radiance, ld_cfs = crv_utils.prep_surface_params(base_container, return_values=True,
                                                                write_to_containers=True, **kwargs)
        if mirror_container is None:
            return normal_radiance, ld_cfs
        for component in settings.BINARY_COUNTERPARTS:
            star = getattr(mirror_container, component)
            setattr(star, 'normal_radiance', normal_radiance[component])
            setattr(star, 'ld_cfs', ld_cfs[component])
        return normal_radiance, ld_cfs
    else:
        for on_pos in [base_container, mirror_container]:
            if on_pos is None:
                continue
            for component in settings.BINARY_COUNTERPARTS:
                star = getattr(on_pos, component)
                setattr(star, 'normal_radiance', old_normal_radiance[component])
                setattr(star, 'ld_cfs', old_ld_cfs[component])
        return old_normal_radiance, old_ld_cfs


def integrate_eccentric_curve_w_orbital_symmetry(*args):
    """
    Curve generator for eccentric curves without spots for couples of orbital positions that
    are symmetrically positioned around apsidal line and thus share the same surface geometry.

    :param args: Tuple;

    ::

        Tuple[
                binary: elisa.binary_system.BinarySystem,
                initial_system: elisa.
                orbital_positions: numpy.array; (N*2*5) stacked couples of orbital positions
                radii: numpy.array; forward radii
                crv_labels: List;
                curves_fn: function to calculate curve points at given orbital positions,
                kwargs: Dict,
            ]

    :return: Dict; curves
    """
    binary, all_potentials, orbital_positions, radii, crv_labels, curve_fn, kwargs = args

    # surface potentials with constant volume of components
    potentials = {
        component: pot[orbital_positions[:, 0, 0].astype(INT)]
        for component, pot in all_potentials.items()
    }

    base_radii = radii[:, orbital_positions[:, 0, 0].astype(INT)]
    rel_d_radii = crv_utils.compute_rel_d_geometry(binary, base_radii[:, 1:], base_radii[:, :-1])
    args = (binary.has_spots(), orbital_positions.shape[0], rel_d_radii)
    new_geometry_mask = dynamic.resolve_object_geometry_update(*args)

    rel_irrad = crv_utils.compute_rel_d_irradiation(binary, orbital_positions[:, 0, 1])
    new_irrad_mask = dynamic.resolve_irrad_update(rel_irrad, orbital_positions.shape[0])

    new_build_mask = np.logical_or(new_geometry_mask, new_irrad_mask)

    curves_body = {key: np.zeros(orbital_positions.shape[0]) for key in crv_labels}
    curves_mirror = {key: np.zeros(orbital_positions.shape[0]) for key in crv_labels}

    # prepare initial orbital position container
    from_this = dict(binary_system=binary, position=const.Position(0, 1.0, 0.0, 0.0, 0.0))
    initial_system = OrbitalPositionContainer.from_binary_system(**from_this)

    ld_cfs, normal_radiance = None, None
    for idx in range(orbital_positions.shape[0]):
        body, mirror = orbital_positions[idx, 0, :], orbital_positions[idx, 1, :]
        base_orb_pos, mirror_orb_pos = utils.convert_binary_orbital_motion_arr_to_positions([body, mirror])

        initial_system.set_on_position_params(base_orb_pos, potentials['primary'][idx], potentials['secondary'][idx])
        initial_system = _update_surface_in_ecc_orbits(initial_system, base_orb_pos, new_build_mask[idx])

        on_pos_base = bsutils.move_sys_onpos(initial_system, base_orb_pos, on_copy=True)
        _kwargs = dict(in_eclipse=True, return_values=False, write_to_containers=True)
        compute_surface_coverage(on_pos_base, binary.semi_major_axis, **_kwargs)

        if OrbitalSupplements.is_empty(mirror):
            on_pos_mirror = None
        else:
            # orbital velocities are not symmetrical along apsidal lines
            d_distance = mirror_orb_pos.distance - base_orb_pos.distance
            initial_system.secondary.points[:, 0] += d_distance
            _kwargs = dict(recalculate_velocities=True, on_copy=True)
            on_pos_mirror = bsutils.move_sys_onpos(initial_system, mirror_orb_pos, **_kwargs)
            _kwargs = dict(in_eclipse=True, return_values=False, write_to_containers=True)
            compute_surface_coverage(on_pos_mirror, binary.semi_major_axis, **_kwargs)

        # normal radiances and ld coefficients will be used for both base and mirror orbital positions
        args = (new_build_mask[idx], on_pos_base, on_pos_mirror, normal_radiance, ld_cfs)
        normal_radiance, ld_cfs = _update_ldc_and_radiance_on_orb_pair(*args, **kwargs)

        curves_body = curve_fn(curves_body, idx, crv_labels, on_pos_base)
        curves_mirror = curves_mirror if on_pos_mirror is None else \
            curve_fn(curves_mirror, idx, crv_labels, on_pos_mirror)

    return {key: np.stack((curves_body[key], curves_mirror[key]), axis=1) for key in crv_labels}


def similar_neighbour_approximation_ecc_curve_integration(*args):
    """
    Curve generator for eccentric curves without spots for orbital positions that are sufficiently similar that surfaces
    does not have to be fully recalculated.

    :param args: Tuple;

    ::

        Tuple[
                binary: elisa.binary_system.BinarySystem,
                potentials: Dict; corrected surface potentials
                motion_batch: numpy.array; orbital positions sorted by components distance
                new_geometry_mask: bool; mask to `orbital_positions` which determines which surface
                                         geometry should be fully recalculated
                crv_labels: List; curve_labels
                curves_fn: function to calculate curve points at given orbital positions,
                kwargs: Dict,
            ]

    :return: Dict; curves
    """
    binary, potentials, motion_batch, new_geometry_mask, crv_labels, curve_fn, kwargs = args
    curves = {key: np.empty(len(motion_batch)) for key in crv_labels}
    positions = utils.convert_binary_orbital_motion_arr_to_positions(motion_batch)

    # prepare initial orbital position container
    from_this = dict(binary_system=binary, position=const.Position(0, 1.0, 0.0, 0.0, 0.0))
    initial_system = OrbitalPositionContainer.from_binary_system(**from_this)

    normal_radiance, ld_cfs = None, None
    for run_idx, position in enumerate(positions):
        pos_idx = int(position.idx)
        require_rebuild = new_geometry_mask[pos_idx] or run_idx == 0

        initial_system.set_on_position_params(position, potentials["primary"][pos_idx],
                                              potentials["secondary"][pos_idx])

        _update_surface_in_ecc_orbits(initial_system, orbital_position=position, new_geometry_test=require_rebuild)
        on_pos = bsutils.move_sys_onpos(initial_system, position, on_copy=True, recalculate_velocities=True)
        on_pos, normal_radiance, ld_cfs = crv_utils.update_surface_params(require_rebuild, on_pos, normal_radiance,
                                                                          ld_cfs, **kwargs)
        # TODO: properly calculate in_eclipse parameter
        compute_surface_coverage(on_pos, binary.semi_major_axis, in_eclipse=True, return_values=False,
                                 write_to_containers=True)
        curves = curve_fn(curves, run_idx, crv_labels, on_pos)
    return curves



import numpy as np

from .. import const as c, units as u
from .. base.transform import SystemProperties, quantity_transform, deg_transform, WHEN_FLOAT64
from .. base.types import INT, FLOAT


class PulsationModeProperties(SystemProperties):
    @staticmethod
    def l(value):
        """
        Returns angular degree of the pulsation mode.

        :param value: int;
        :return: int; validated value of angular degree
        """
        if int(value) - value == 0:
            value = int(value)
        if not isinstance(value, (int, INT)):
            raise TypeError('Angular degree `l` is not (numpy.)int ')
        return value

    @staticmethod
    def m(value):
        """
        Returns azimuthal order of the pulsation mode.

        :param value: int;
        :return: int; validated value of azimuthal order
        """
        if int(value) - value == 0:
            value = int(value)
        if not isinstance(value, (int, INT)):
            raise TypeError('Angular degree `m` is not (numpy.)int ')
        return value

    @staticmethod
    def amplitude(value):
        """
        Returns evaluated radial velocity amplitude of pulsation mode.

        :param value: Union[float, astropy.unit.quantity.units.Quantity]
        :return: float;
        """
        if isinstance(value, (u.Quantity, str)):
            retval = u.Quantity(value) if isinstance(value, str) else value
            retval = np.float64(retval.to(u.VELOCITY_UNIT))
        elif isinstance(value, (int, INT, float, FLOAT)):
            retval = FLOAT(value)
        else:
            raise TypeError('Value of `amplitude` is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        if retval < 0:
            raise ValueError('Temperature amplitude of mode has to be non-negative number.')

        return retval

    @staticmethod
    def frequency(value):
        """
        Returns evaluated frequency of pulsation mode.

        :param value: Union[float, astropy.unit.quantity.units.Quantity]
        :return: float;
        """
        if isinstance(value, (u.Quantity, str)):
            retval = u.Quantity(value) if isinstance(value, str) else value
            retval = np.float64(retval.to(u.FREQUENCY_UNIT))
        elif isinstance(value, (int, INT, float, FLOAT)):
            retval = (FLOAT(value) * u.DefaultPulsationsInputUnits['frequency']).to(u.FREQUENCY_UNIT).value
        else:
            raise TypeError('Value of `frequency` is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        if retval < 0:
            raise ValueError('Frequency of the mode has to be non-negative number.')

        return retval

    @staticmethod
    def start_phase(value):
        """
        Returns phase shift of the given pulsation mode.

        :param value: float;
        :return: float;
        """
        return quantity_transform(value, u.DefaultPulsationsUnits.start_phase, WHEN_FLOAT64,
                                  u.DefaultPulsationsInputUnits.start_phase)

    @staticmethod
    def mode_axis_theta(value):
        """
        Evaluates value for latitude of pulsation mode axis.
        If unit is not supplied, degrees are assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.units.units.Quantity]
        :return: float;
        """
        retval = deg_transform(value, u.DefaultPulsationsUnits.mode_axis_theta, WHEN_FLOAT64,
                               u.DefaultPulsationsInputUnits.mode_axis_theta)
        if not 0 <= retval < c.PI:
            raise ValueError(f'Value of `mode_axis_theta`: {retval} is outside bounds (0, pi).')

        return retval

    @staticmethod
    def mode_axis_phi(value):
        """
        Evaluates longitude of pulsation mode axis.
        If unit is not supplied, degrees are assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float;
        """
        return deg_transform(value, u.DefaultPulsationsUnits.mode_axis_phi, WHEN_FLOAT64,
                             u.DefaultPulsationsInputUnits.mode_axis_phi)

    @staticmethod
    def temperature_perturbation_phase_shift(value):
        """
        evaluates a phase shift between surface geometry perturbation and temperature perturbations

        :param value: float; phase shift in radians
        :return:
        """
        return deg_transform(value, u.DefaultPulsationsUnits.temperature_perturbation_phase_shift, WHEN_FLOAT64,
                             u.DefaultPulsationsInputUnits.temperature_perturbation_phase_shift)

    @staticmethod
    def horizontal_to_radial_amplitude_ratio(value):
        """
        Defines amplitude ratio between radial and horizontal displacement.

        :param value: float; phase shift in radians
        :return:
        """
        if not isinstance(value, (int, INT, float, FLOAT)):
            raise TypeError('Parameter is not is not (numpy.)int or (numpy.)float')
        return value

    @staticmethod
    def tidally_locked(value):
        """
        Defines whether the pulsation mode is fixed with respect to the tidal axis.
        If false, the mode axis will drift with the stellar surface.

        :param value: bool;
        :return: bool
        """
        if not isinstance(value, bool):
            raise TypeError('Parameter `tidally_locked` can contain only boolean')
        return value



import numpy as np

from . import model
from .. import const
from .. opt.fsolver import fsolve


def calculate_radius(mass, angular_velocity, surface_potential, *args):
    """
    Function calculates radius of the star in given direction of arbitrary direction vector (in spherical
    coordinates) starting from the centre of the star.

    :param mass: float;
    :param angular_velocity: float;
    :param surface_potential: float;
    :param args: Tuple;

    ::

        (
            theta: float - latitudonal angle of direction vector measured from north pole (in radians)
        )

    :return: float; radius
    """
    fn = model.potential_fn
    precalc = model.pre_calculate_for_potential_value

    precalc_args = (mass, angular_velocity) + args
    init_val = - const.G * mass / surface_potential
    scipy_solver_init_value = np.array([init_val])
    argss = (precalc(*precalc_args), surface_potential)
    solution, _, ier, _ = fsolve(fn, scipy_solver_init_value, full_output=True, args=argss, xtol=1e-10)
    # check for regular solution
    if ier == 1 and not np.isnan(solution[0]) and 5 * init_val >= solution[0] >= 0:
        return solution[0]
    else:
        if not (0 < solution[0] < 5 * init_val):
            raise ValueError(f'Invalid value of radius {solution} was calculated.')
        return solution[0]


def calculate_polar_radius(mass, angular_velocity, surface_potential):
    args = (0.0, )
    return calculate_radius(mass, angular_velocity, surface_potential, *args)


def calculate_equatorial_radius(mass, angular_velocity, surface_potential):
    args = (const.HALF_PI, )
    return calculate_radius(mass, angular_velocity, surface_potential, *args)



import numpy as np

from . import utils as putils
from . surface import kinematics
from . import pulsations
from .. import utils, const
from .. base.surface.faces import (
    set_all_surface_centres,
    calculate_normals
)
from .. base.types import COMPLEX


def generate_harmonics(star_container, com_x, phase, time):
    """
    Generating spherical harmonics Y_l^m in shapes(2, n_points) and (2, n_faces)
    and its derivatives to be subsequently used for calculation of perturbed properties.

    :param star_container: elisa.base.container.StarContainer;
    :param com_x: float; centre of mass for the component
    :param phase: float; rotational/orbital phase
    :param time: float; time of the observation
    :return: elisa.base.container.StarContainer; Star container with updated harmonics
    """
    if not star_container.is_flat():
        raise ValueError('Pulsations can be calculated only on flattened container.')
    _kwargs = dict(kind='points', com_x=com_x)
    star_container.points_spherical = star_container.transform_points_to_spherical_coordinates(**_kwargs)

    tilt_phi, tilt_theta = putils.generate_tilt_coordinates(star_container, phase)
    star_container.pulsations[0].tilt_phi, star_container.pulsations[0].tilt_theta = tilt_phi, tilt_theta

    tilted_points = putils.tilt_mode_coordinates(
        star_container.points_spherical, star_container.pulsations[0].tilt_phi, star_container.pulsations[0].tilt_theta
    )

    # assigning tilted points in spherical coordinates only to the first mode (the rest will share the same points)
    star_container.pulsations[0].points = tilted_points

    exponential, norm_constant = dict(), dict()
    for mode_index, mode in star_container.pulsations.items():
        # beware of this in case you want to use the container at different phase
        exponential[mode_index] = putils.generate_time_exponential(mode, time)

        # generating harmonics Y_m^l and Y_m+1^l for star and spot points
        harmonics = np.zeros((2, tilted_points.shape[0]), dtype=COMPLEX)
        harmonics[0] = pulsations.spherical_harmonics(mode, tilted_points, exponential[mode_index])
        if mode.m != mode.l:
            _args, _kwargs = (mode, tilted_points, exponential[mode_index]), dict(order=mode.m + 1, degree=mode.l)
            harmonics[1] = pulsations.spherical_harmonics(*_args, **_kwargs)

        # generating derivatives of spherical harmonics by phi an theta
        derivatives = np.empty((2, tilted_points.shape[0]), dtype=COMPLEX)
        derivatives[0] = pulsations.diff_spherical_harmonics_by_phi(mode, harmonics)
        _args = (mode, harmonics, tilted_points[:, 1], tilted_points[:, 2])
        derivatives[1] = pulsations.diff_spherical_harmonics_by_theta(*_args)

        # renormalizing horizontal amplitude to 1
        if mode.l > 0:
            norm_constant[mode_index] = pulsations.horizontal_displacement_normalization(derivatives, harmonics)
            derivatives *= norm_constant[mode_index]

        # assignment of harmonics to mode instance variables
        mode.point_harmonics = harmonics[0]
        mode.point_harmonics_derivatives = derivatives

    return star_container


def incorporate_pulsations_to_model(star_container, com_x, scale=1.0):
    """
    Function adds perturbation to the surface mesh due to pulsations.

    :param star_container: base.container.StarContainer;
    :param com_x: float;
    :param scale: numpy.float; scale of the perturbations
    :return: base.container.StarContainer;
    """
    # calculating kinematics quantities
    complex_displacement(star_container, scale=scale)

    # treating polar regions
    putils.pole_neighbours(star_container)

    position_perturbation(star_container, com_x=com_x, update_container=True, return_perturbation=False)
    velocity_perturbation(star_container, scale=scale, update_container=True, return_perturbation=False)
    gravity_acc_perturbation(star_container, scale=scale, update_container=True, return_perturbation=False)
    temp_perturbation(star_container, scale=scale, update_container=True, return_perturbation=False)

    # recalculating normals and areas
    set_all_surface_centres(star_container)
    args = (star_container.points, star_container.faces, star_container.face_centres, com_x)
    star_container.normals = calculate_normals(*args)
    star_container.calculate_all_areas()
    return star_container


def complex_displacement(star, scale):
    """
    Assigning complex displacement for surface points. Complex displacement
    is then used to calculate the kinematic quantities (r, v, a).

    :param star: base.container.StarContainer;
    :param scale: float;
    :return: base.container.StarContainer;
    """
    if not star.is_flat():
        raise ValueError('Pulsations can be calculated only on flattened container.')

    for mode_index, mode in star.pulsations.items():
        mode.complex_displacement = kinematics.calculate_displacement_coordinates(
            mode, star.pulsations[0].points, mode.point_harmonics, mode.point_harmonics_derivatives,
            star.points_spherical[:, 0], scale=scale
        )

    return star


def position_perturbation(
        star,
        com_x=0,
        update_container=True,
        return_perturbation=False,
        spherical_perturbation=False
):
    """
    Calculates the deformation of the surface mesh due to the pulsations.

    :param com_x: float; x coordinate of the compotnents centre of mass in corotating frame
    :param star: base.container.StarContainer;
    :param update_container: bool; if True, perturbation is incorporated into star.points
    :param return_perturbation: bool; if True, calculated displacement (in cartesian coordinates) is returned
    :param spherical_perturbation: bool; if True, perturbations are returned in spherical coordinates
    :return: Union[numpy.array, None];
    """
    displacement = None

    tilt_displacement_sph = np.sum([
        kinematics.calculate_mode_angular_displacement(mode.complex_displacement) for mode in star.pulsations.values()
    ], axis=0)

    points_spherical = putils.derotate_surface_points(
        star.pulsations[0].points + tilt_displacement_sph,
        star.pulsations[0].tilt_phi, star.pulsations[0].tilt_theta
    )

    points = utils.spherical_to_cartesian(points_spherical)
    if return_perturbation:
        if spherical_perturbation:
            displacement = points_spherical - getattr(star, 'points_spherical')
            displacement[displacement[:, 1] > const.PI, 1] -= const.FULL_ARC
        else:
            points - utils.spherical_to_cartesian(star.points_spherical)
    if update_container:
        com = np.array([com_x, 0, 0])
        setattr(star, 'points', points + com[None, :])

    return displacement if return_perturbation else None


def velocity_perturbation(
        star,
        scale,
        update_container=False,
        return_perturbation=False,
        spherical_perturbation=False,
        point_perturbations=False
):
    """
    Calculates velocity perturbation on a surface of a pulsating star.

    :param star: base.container.StarContainer;
    :param scale: float; scaling factor of the system (a in case of BinarySystem)
    :param update_container: bool; if true, the perturbations are added into surface element velocities
    :param return_perturbation: bool; if True, velocity perturbation itself is returned
    :param spherical_perturbation: bool; if True, velocity perturbation in spherical coordinates (d_r, d_phi, d_theta)
                                         is returned.
    :param point_perturbations: bool; if True perturbations on surface pints are returned instead face averaged values
    :return: Union[None, numpy.array];
    """
    # calculating perturbed velocity in spherical coordinates
    tilt_velocity_sph = np.sum([
        kinematics.calculate_mode_derivatives(
            displacement=mode.complex_displacement, angular_frequency=mode.angular_frequency
        ) for mode in star.pulsations.values()
    ], axis=0)

    velocity_pert_sph = putils.derotate_surface_displacements(
        tilt_velocity_sph, star.pulsations[0].points, star.points_spherical,
        star.pulsations[0].tilt_phi, star.pulsations[0].tilt_theta
    )
    velocity_pert_sph[star.pole_idx] = velocity_pert_sph[star.pole_idx_neighbour]
    points_cartesian = utils.spherical_to_cartesian(star.points_spherical)
    args = (velocity_pert_sph, points_cartesian, 0.0)
    velocity_pert = putils.transform_spherical_displacement_to_cartesian(*args) * scale

    velocity_pert_face = velocity_pert[star.faces].mean(axis=1)

    if update_container:
        star.velocities += velocity_pert_face

    if return_perturbation:
        if spherical_perturbation:
            velocity_pert_sph[:, 0] *= scale
            return velocity_pert_sph[star.faces].mean(axis=1) if not point_perturbations else velocity_pert_sph
        else:
            return velocity_pert if point_perturbations else velocity_pert_face
    else:
        return None


def gravity_acc_perturbation(
        star,
        scale,
        update_container=False,
        return_perturbation=False,
        spherical_perturbation=False,
        point_perturbations=False
):
    """
    Calculates acceleration perturbation on a surface of a pulsating star.

    :param star: base.container.StarContainer;
    :param scale: float; scaling factor of the system (a in case of BinarySystem)
    :param update_container: bool; if true, the perturbations are added into surface element velocities
    :param return_perturbation: bool; if True, velocity perturbation itself is returned
    :param spherical_perturbation: bool; if True, velocity perturbation in spherical coordinates (d_r, d_phi, d_theta)
                                         is returned.
    :param point_perturbations: bool; if True perturbations on surface pints are returned instead face averaged values
    :return:
    """
    # calculating perturbed acceleration in tilted spherical coordinates
    tilt_acc_sph = np.sum([
        kinematics.calculate_mode_second_derivatives(
            displacement=mode.complex_displacement, angular_frequency=mode.angular_frequency
        ) for mode in star.pulsations.values()
    ], axis=0)

    acc_pert_sph = putils.derotate_surface_displacements(
        tilt_acc_sph, star.pulsations[0].points, star.points_spherical,
        star.pulsations[0].tilt_phi, star.pulsations[0].tilt_theta
    )
    acc_pert_sph[star.pole_idx] = acc_pert_sph[star.pole_idx_neighbour]
    points_cartesian = utils.spherical_to_cartesian(star.points_spherical)
    acc_pert = putils.transform_spherical_displacement_to_cartesian(acc_pert_sph, points_cartesian, 0.0) * scale

    # treating singularities at poles
    acc_pert[star.pole_idx] = acc_pert[star.pole_idx_neighbour]

    acc_pert_face = acc_pert[star.faces].mean(axis=1)

    if update_container:
        g_eq = - np.power(10, star.log_g)[:, None] * star.normals
        total_acc = np.linalg.norm(g_eq + acc_pert_face, axis=1)
        star.log_g = np.log10(total_acc)

    if return_perturbation:
        if spherical_perturbation:
            acc_pert_sph[:, 0] *= scale
            return acc_pert_sph[star.faces].mean(axis=1) if not point_perturbations else acc_pert_sph
        else:
            return acc_pert if point_perturbations else acc_pert_face
    else:
        return None


def temp_perturbation(star, scale=1.0, update_container=False, return_perturbation=False, point_perturbations=False):
    """
    Calculates temperature perturbation on a surface of a pulsating star.

    :param star: base.container.StarContainer;
    :param scale: numpy.float; scale of the system
    :param update_container: bool; if true, the perturbations are added into surface element temperatures
    :param return_perturbation:
    :param point_perturbations: bool; if True perturbations on surface pints are returned instead face averaged values
    :return:
    """
    temp_pert = np.sum([kinematics.calculate_temperature_pert_factor(mode, scale)
                        for mode in star.pulsations.values()], axis=0)

    temp_pert_face = temp_pert[star.faces].mean(axis=1) * star.temperatures
    if update_container:
        star.temperatures += temp_pert_face

    if return_perturbation:
        return temp_pert*star.t_eff if point_perturbations else temp_pert_face



import numpy as np
from packaging import version

from . types import INT, FLOAT
from .. import (
    units as u,
    const,
    settings
)
from .. units import (
    DefaultStarInputUnits,
    DefaultBinarySystemInputUnits,
    DefaultSystemInputUnits,
    DefaultSystemUnits
)


WHEN_FLOAT64 = (int, np.int32, np.int64, float, np.float32, np.float64)
if version.parse(np.__version__) < version.parse("1.20.0"):
    # noinspection PyUnresolvedReferences
    WHEN_FLOAT64 += (np.int, np.float)

WHEN_ARRAY = (list, np.ndarray, tuple)


def quantity_transform(value, unit, when_float64=WHEN_FLOAT64, default_input_unit=None):
    """
    General transform function for quantities which fit such interface.

    :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity, str]; string accepts the string
                                                                                           representation of the astropy
                                                                                           units
    :param unit: astropy.units.Quantity; unit of the output
    :param when_float64: Tuple(Types)
    :param default_input_unit: astropy.units.Unit; assumed unit of input it unit is not supplied in `value`
    :return: float
    """
    if isinstance(value, (u.Quantity, str)):
        value = u.Quantity(value) if isinstance(value, str) else value
        value = np.float64(value.to(unit))
    elif isinstance(value, when_float64):
        value = np.float64(value) if default_input_unit is None \
            else (value * default_input_unit).to(unit).value
    else:
        raise TypeError('Input of variable is not (numpy.)int or (numpy.)float '
                        'nor astropy.unit.quantity.Quantity instance (or its string representation).')
    return value


def deg_transform(value, unit, when_float64, default_input_unit=u.deg):
    """
    General transform function for angular quantities which fit such interface.

    :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity, str]; string accepts the string
                                                                                           representation of the astropy
                                                                                           units
    :param unit: astropy.units.Quantity; unit of the output
    :param when_float64:
    :param default_input_unit: astropy.units.Unit; assumed unit of input it unit is not supplied in `value`
    :return: float;
    """
    if isinstance(value, (u.Quantity, str)):
        value = u.Quantity(value) if isinstance(value, str) else value
        value = np.float64(value.to(unit))
    elif isinstance(value, when_float64):
        value = np.float64(value)*default_input_unit.to(unit)
    else:
        raise TypeError('Input of the angular variable is not (numpy.)int or (numpy.)float '
                        'nor astropy.unit.quantity.Quantity instance (or its string representation).')
    return value


class TransformProperties(object):
    @classmethod
    def transform_input(cls, **kwargs):
        """
        Function transforms input dictionary of keyword arguments of the System to internally usable state
        (conversion and stripping of units).

        :param kwargs: Dict;
        :return: Dict;
        """
        return {key: getattr(cls, key)(val) if hasattr(cls, key) else val for key, val in kwargs.items()}


class SystemProperties(TransformProperties):

    @staticmethod
    def inclination(value):
        """
        Transform and validity check for inclination of the system.
        If unit is not supplied, value in degrees is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity, str]
        :return:
        """
        if isinstance(value, (u.Quantity, str)):
            value = u.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(u.DefaultSystemUnits.inclination))
        elif isinstance(value, (int, INT, float, FLOAT)):
            value = np.float64((value * DefaultSystemInputUnits.inclination).to(DefaultSystemUnits.inclination))
        else:
            raise TypeError('Input of variable `inclination` is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance (or its string representation).')

        if not 0 <= value <= const.PI:
            raise ValueError(f'Inclination value of {value} is out of bounds (0, pi).')
        return value

    @staticmethod
    def period(value):
        """
        Transform and validate orbital period of binary star system.
        If unit is not specified, default period unit is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return quantity_transform(value, DefaultSystemUnits.period, WHEN_FLOAT64, u.DefaultSystemInputUnits.period)

    @staticmethod
    def gamma(value):
        """
        Validate and transform center of mass velocity.
        Expected type is astropy.units.quantity.Quantity, numpy.float or numpy.int othervise TypeError will be raised.
        If unit is not specified, default velocity unit is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return:
        """
        return quantity_transform(value, u.DefaultSystemUnits.gamma, WHEN_FLOAT64, u.DefaultSystemInputUnits.gamma)

    @staticmethod
    def additional_light(value):
        """
        Validate and transform for additional light - light that does not originate from any member of system.

        :param value: float (0, 1)
        :return:
        """
        if not 0.0 <= value <= 1.0:
            raise ValueError('Invalid value of additional light. Valid values are between 0 and 1.')
        return np.float64(value)

    @staticmethod
    def semi_major_axis(value):
        """
        Validate and transform for semi major axis.

        :param value:
        :return:
        """
        if isinstance(value, (u.Quantity, str)):
            value = u.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(u.DefaultBinarySystemUnits.system.semi_major_axis))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value * DefaultBinarySystemInputUnits.system.semi_major_axis.
                               to(u.DefaultBinarySystemUnits.system.semi_major_axis))
        else:
            raise TypeError('User input is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance (or its string representation).')
        if value <= 0:
            raise ValueError("Invalid value of semi_major_axis, use value > 0!")
        return value

    @staticmethod
    def distance(value):
        if isinstance(value, (u.Quantity, str)):
            value = u.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(u.DefaultSystemUnits.distance))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value * DefaultBinarySystemInputUnits.system.distance.to(u.DefaultSystemUnits.distance))
        else:
            raise TypeError('User input is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance (or its string representation).')
        if value <= 0:
            raise ValueError("Invalid value of system`s distance, use value > 0!")

        return value


class BodyProperties(TransformProperties):
    @staticmethod
    def synchronicity(value):
        """
        Object synchronicity (F = omega_rot/omega_orb) validator.
        Expects number input convertible to numpy float64 / float.

        :param value: float
        """
        if value <= 0:
            raise ValueError("Invalid synchronicity, use value > 0!")
        return np.float64(value)

    @staticmethod
    def albedo(value):
        """
        Validate and transform bolometric albedo (reradiated energy/ irradiance energy).
        Accepts value of albedo in range (0, 1).

        :param value: float;
        :return: float;
        """
        if value < 0 or value > 1:
            raise ValueError(f'Parameter albedo = {value} is out of range <0, 1>')
        return np.float64(value)

    @staticmethod
    def discretization_factor(value):
        """
        Discretization factor. Degrees is considered as default value.

        :param value: Union[float, astropy.quantity.Quantity]
        :return: float
        """
        value = deg_transform(value, u.DefaultStarUnits.discretization_factor, WHEN_FLOAT64,
                              u.DefaultStarInputUnits.discretization_factor)
        if value > const.HALF_PI:
            raise ValueError("Invalid value of alpha parameter. Use value less than 90.")
        return value

    @staticmethod
    def t_eff(value):
        """
        This function accepts value in Any temperature unit.
        If your input is without unit, function assumes that supplied value is in Kelvins.

        :param value: Union[int, numpy.int, float, numpy.float, astropy.unit.quantity.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultStarUnits.t_eff, WHEN_FLOAT64, u.DefaultStarInputUnits.t_eff)


class StarProperties(BodyProperties):
    @staticmethod
    def equivalent_radius(value):
        """
        Expected type is astropy.units.quantity.Quantity, numpy.float or numpy.int othervise TypeError will be raised.
        If quantity is not specified, default distance unit is assumed.

        :param value: float
        :return:
        """
        if isinstance(value, (u.Quantity, str)):
            value = u.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(u.DefaultStarUnits.equivalent_radius))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value * DefaultStarInputUnits.equivalent_radius.to(u.DefaultStarUnits.equivalent_radius))
        else:
            raise TypeError('User input is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance (or its string representation).')
        if value <= 0:
            raise ValueError("Invalid value of equivalent_radius, use value > 0!")
        return value

    @staticmethod
    def mass(value):
        """
       If mass is int, np.int, float, np.float, program assumes solar mass as it's unit.
       If mass astropy.unit.quantity.Quantity instance, program converts it to default units and stores it's value in
       attribute _mass.

       :param value: Union[int, numpy.int, float, numpy.float, astropy.unit.quantity.Quantity, str];
       :return: float
       """
        if isinstance(value, (u.Quantity, str)):
            value = u.Quantity(value) if isinstance(value, str) else value
            value = np.float64(value.to(u.DefaultStarUnits.mass))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value * DefaultStarInputUnits.mass.to(u.DefaultStarUnits.mass))
        else:
            raise TypeError('User input is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance (or its string representation).')
        if value <= 0:
            raise ValueError("Invalid mass, use value > 0!")
        return value

    @staticmethod
    def surface_potential(value):
        """
        Returns surface potential of Star.

        :param value: float
        :return: float
        """
        if value < 0:
            raise ValueError("Invalid surface potential, use value > 0!")
        return np.float64(value)

    @staticmethod
    def metallicity(value):
        if not isinstance(value, WHEN_FLOAT64):
            raise TypeError('Input of variable `metallicity` is not (np.)int or (np.)float instance.')
        return np.float64(value)

    @staticmethod
    def polar_log_g(value):
        """
        Setter for polar surface gravity.
        If unit is not specified in astropy.units format, value in cgs unit is assumed (it means log(g) in cgs).

        :param value: float or astropy.unit.quantity.Quantity
        :return: float
        """
        if isinstance(value, (u.Quantity, str)):
            value = u.Dex(value, unit=u.Unit(' '.join(value.split()[1:]))) if isinstance(value, str) else value
            value = np.float64(value.to(u.LOG_ACCELERATION_UNIT))
        elif isinstance(value, WHEN_FLOAT64):
            # conversion from cgs to SI
            value -= 2
        else:
            raise TypeError('User input is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance (or its string representation).')
        return value

    @staticmethod
    def gravity_darkening(value):
        """
        Gravity darkening validator.

        :param value: float;
        :return: float
        """
        if value > 1 or value < 0:
            raise ValueError(f'Parameter gravity darkening = {value} is out of range <0, 1>')
        return np.float64(value)

    @staticmethod
    def limb_darkening_coefficients(value):
        """
        Validates custom limb darkening coefficients.

        :param value: Union[none, dict];
        :return: Union[none, dict];
        """
        if value is None:
            retval = None  # default case of interpolated LD coefficients
        elif isinstance(value, dict):
            retval = dict()
            for passband, ld_coeffs in value.items():
                if isinstance(ld_coeffs, WHEN_FLOAT64):
                    if settings.LIMB_DARKENING_LAW in ['linear', 'cosine']:
                        retval[passband] = [ld_coeffs, ]
                    else:
                        raise TypeError('Scalar limb darkening coefficient is available only for linear (cosine) law.')
                elif isinstance(ld_coeffs, WHEN_ARRAY):
                    desired_vector_length = len(settings.LD_LAW_CFS_COLUMNS[settings.LIMB_DARKENING_LAW])
                    if np.shape(ld_coeffs)[0] != desired_vector_length:
                        raise ValueError(f'{settings.LIMB_DARKENING_LAW} limb-darkening law requires '
                                         f'{desired_vector_length} components in a vector with shape '
                                         f'({desired_vector_length}, ), however, you provided a vector with '
                                         f'{len(ld_coeffs)} components with shape {np.shape(ld_coeffs)}')
                    else:
                        retval[passband] = ld_coeffs
        else:
            raise TypeError('Limb darkening coefficients needs to be supplied as dictionary in format '
                            '{`passband`: ld_coeffs, }, where `ld_coeffs` is float (for linear law) '
                            'and/or list (numpy.array) in other cases.')

        return retval


class SpotProperties(BodyProperties):
    @staticmethod
    def latitude(value):
        """
        Expecting value in degrees or as astropy units instance.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return deg_transform(value, u.DefaultSpotUnits.latitude, WHEN_FLOAT64, u.DefaultSpotInputUnits.latitude)

    @staticmethod
    def longitude(value):
        """
        Expecting value in degrees or as astropy units instance.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return deg_transform(value, u.DefaultSpotUnits.longitude, WHEN_FLOAT64, u.DefaultSpotInputUnits.longitude)

    @staticmethod
    def angular_radius(value):
        """
        Expecting value in degrees or as astropy units instance.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return:
        """
        return deg_transform(value, u.DefaultSpotUnits.angular_radius, WHEN_FLOAT64,
                             u.DefaultSpotInputUnits.angular_radius)

    @staticmethod
    def temperature_factor(value):
        """
        Validate and transform temperature factor.

        :param value: flaot
        :return: float
        """
        if not isinstance(value, (int, INT, float, FLOAT)):
            raise TypeError('Input of variable `temperature_factor` is not (numpy.)int or (numpy.)float.')
        return np.float64(value)

    @staticmethod
    def discretization_factor(value):
        """
        Discretization factor. Degrees is considered as default value.

        :param value: Union[float, astropy.quantity.Quantity]
        :return: float
        """
        value = deg_transform(value, u.DefaultSpotUnits.discretization_factor, WHEN_FLOAT64,
                              u.DefaultSpotInputUnits.discretization_factor)
        if value > const.HALF_PI:
            raise ValueError("Invalid value of alpha parameter. Use value less than 90.")
        return value



# noinspection PyTypeChecker
import numpy as np
import json

from copy import copy
from .. import utils as bsutils
from .. surface import faces as bsfaces
from ... base.types import FLOAT
from ... logger import getLogger
from ... import settings
from ... utils import is_empty
from ... base.surface import temperature as btemperature
from ... import (
    umpy as up,
    ld,
    utils
)
from elisa.base.surface.temperature import renormalize_temperatures
from elisa.numba_functions import reflection_effect as re_numba


logger = getLogger("binary_system.surface.temperature")


def redistribute_temperatures(in_system, temperatures):
    """
    In this function array of `temperatures` is parsed into chunks that belong to stellar surface and spots.

    :param in_system: instance to redistribute temperatures
    :param temperatures: numpy.array; temperatures from the whole surface, ordered: surface, spot1, spot2...
    :param in_system: instance to redistribute temperatures
    """
    for component in ['primary', 'secondary']:
        star = getattr(in_system, component)
        counter = len(star.temperatures)
        star.temperatures = temperatures[component][:counter]
        if star.has_spots():
            for spot_index, spot in star.spots.items():
                spot.temperatures = temperatures[component][counter: counter + len(spot.temperatures)]
                counter += len(spot.temperatures)
    return in_system


def apply_reflection_effect(system, components_distance, iterations):
    """
    Alter temperatures of components to involve reflection effect.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param iterations: int; iterations of reflection effect counts
    :param components_distance: float; components distance in SMA units
    :return: system; elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """

    if not settings.REFLECTION_EFFECT:
        logger.debug('reflection effect is switched off')
        return system
    if iterations <= 0:
        logger.debug('number of reflections in reflection effect was set to zero or negative; '
                     'reflection effect will not be calculated')
        return system

    components = bsutils.component_to_list(component='all')

    _args = (system.primary.polar_radius, system.secondary.polar_radius, components_distance)
    xlim = bsfaces.faces_visibility_x_limits(*_args)

    # this tests if you can use surface symmetries
    not_pulsation_test = not system.has_pulsations()
    not_spot_test = not system.has_spots()
    use_quarter_star_test = not_pulsation_test and not_spot_test
    vis_test_symmetry = {}

    # declaring variables
    centres, vis_test, gamma, normals = {}, {}, {}, {}
    faces, points, temperatures, areas, log_g = {}, {}, {}, {}, {}
    # centres - dict with all centres concatenated (star and spot) into one matrix for convenience
    # vis_test - dict with bool map for centres to select only faces visible from any face on companion
    # companion
    # gamma is of dimensions num_of_visible_faces_primary x num_of_visible_faces_secondary

    # selecting faces that have a chance to be visible from other component
    for component in components:
        star = getattr(system, component)

        points[component], faces[component], centres[component], normals[component], temperatures[component], \
            areas[component], log_g[component] = init_surface_variables(star)

        # test for visibility of star faces
        vis_test[component], vis_test_symmetry[component] = bsfaces.get_visibility_tests(centres[component],
                                                                                         use_quarter_star_test,
                                                                                         xlim[component],
                                                                                         component,
                                                                                         system.morphology)
        if star.has_spots():
            # including spots into overall surface
            for spot_index, spot in star.spots.items():
                vis_test_spot = bsfaces.visibility_test(spot.face_centres, xlim[component], component)

                # merge surface and spot face parameters into one variable
                centres[component], normals[component], temperatures[component], areas[component], \
                    vis_test[component], log_g[component] = \
                    include_spot_to_surface_variables(centres[component], spot.face_centres,
                                                      normals[component], spot.normals,
                                                      temperatures[component], spot.temperatures,
                                                      areas[component], spot.areas, log_g[component],
                                                      spot.log_g, vis_test[component], vis_test_spot)

    # limb darkening coefficients for each face of each component
    ldc = {cmp: ld.get_bolometric_ld_coefficients(temperatures[cmp], log_g[cmp],
                                                  getattr(system, cmp).metallicity,
                                                  getattr(system, cmp).limb_darkening_coefficients)
           for cmp in components}

    # calculating C_A = (albedo_A / D_intB) - scalar
    # D_intB - bolometric limb darkening factor
    d_int = {cmp: ld.calculate_integrated_limb_darkening_factor(settings.LIMB_DARKENING_LAW, ldc[cmp])
             for cmp in components}
    _c = {
        'primary': (system.primary.albedo / d_int['primary']),
        'secondary': (system.secondary.albedo / d_int['secondary'])
    }

    # setting reflection factor R = 1 + F_irradiated / F_original, initially equal to one everywhere - vector
    reflection_factor = {cmp: np.ones(np.sum(vis_test[cmp]), dtype=FLOAT) for cmp in components}
    counterpart = settings.BINARY_COUNTERPARTS

    # for faster convergence, reflection effect is calculated first on cooler component
    components = ['primary', 'secondary'] if system.primary.t_eff <= system.secondary.t_eff else \
        ['secondary', 'primary']

    # pre-calculating 4th power of visible teff for visible triangles
    teff4 = {component: np.empty(temperatures[component].shape) for component in components}
    for cmp in components:
        teff4[cmp][vis_test[cmp]] = up.power(temperatures[cmp][vis_test[cmp]], 4)

    if use_quarter_star_test:
        # calculating distances and distance vectors between, join vector is already normalized
        shp, shp_reduced = get_distance_matrix_shape(system, vis_test)

        distance, join_vector = get_symmetrical_distance_matrix(shp, shp_reduced, centres, vis_test, vis_test_symmetry)

        # calculating cos of angle gamma between face normal and join vector
        # initialising gammma matrices
        gamma = get_symmetrical_gammma(shp[:2], shp_reduced, normals, join_vector, vis_test, vis_test_symmetry)

        # testing mutual visibility of faces by assigning 0 to non visible face combination
        check_symmetric_gamma_for_negative_num(gamma, shp_reduced)

        # calculating QAB = (cos gamma_a)*cos(gamma_b)/d**2
        q_ab = get_symmetrical_q_ab(shp[:2], shp_reduced, gamma, distance)

        # calculating limb darkening factor for each combination of surface faces
        d_gamma = get_symmetrical_d_gamma(shp[:2], shp_reduced, ldc, gamma)

        # calculating limb darkening factors for each combination of faces shape
        # (N_faces_primary * N_faces_secondary)
        # precalculating matrix part of reflection effect correction
        matrix_to_sum2 = {
            'primary': q_ab[:shp_reduced[0], :] * d_gamma['secondary'][:shp_reduced[0], :],
            'secondary': q_ab[:, :shp_reduced[1]] * d_gamma['primary'][:, :shp_reduced[1]]
        }
        symmetry_to_use = {'primary': shp_reduced[0], 'secondary': shp_reduced[1]}
        for _ in range(iterations):
            for component in components:
                star = getattr(system, component)
                counterpart = 'primary' if component == 'secondary' else 'secondary'

                # calculation of reflection effect correction as
                # 1 + (c / t_effi) * sum_j(r_j * Q_ab * t_effj^4 * D(gamma_j) * areas_j)
                # calculating vector part of reflection effect correction
                key_ = counterpart
                vector_to_sum1 = reflection_factor[key_] * teff4[key_][vis_test[key_]] * areas[key_][vis_test[key_]]
                counterpart_to_sum = up.matmul(vector_to_sum1, matrix_to_sum2['secondary']) \
                    if component == 'secondary' else up.matmul(matrix_to_sum2['primary'], vector_to_sum1)
                reflection_factor[component][:symmetry_to_use[component]] = \
                    1 + (_c[component][vis_test_symmetry[component]] / teff4[component][vis_test_symmetry[component]]) \
                    * counterpart_to_sum

                # using symmetry to redistribute reflection factor R
                refl_fact_aux = np.empty(shape=np.shape(temperatures[component]))
                refl_fact_aux[vis_test_symmetry[component]] = reflection_factor[component][:symmetry_to_use[component]]
                refl_fact_aux = star.mirror_face_values(refl_fact_aux)
                reflection_factor[component] = refl_fact_aux[vis_test[component]]

        for component in components:
            star = getattr(system, component)
            # assigning new temperatures according to last iteration as
            # teff_new = teff_old * reflection_factor^0.25
            temperatures[component][vis_test_symmetry[component]] = \
                temperatures[component][vis_test_symmetry[component]] * \
                up.power(reflection_factor[component][:symmetry_to_use[component]], 0.25)
            temperatures[component] = star.mirror_face_values(temperatures[component])

    else:
        # calculating distances and distance vectors between, join vector is already normalized
        distance, join_vector = utils.calculate_distance_matrix(points1=centres['primary'][vis_test['primary']],
                                                                points2=centres['secondary'][vis_test['secondary']],
                                                                return_join_vector_matrix=True)

        # calculating cos of angle gamma between face normal and join vector
        gamma = \
            {'primary': re_numba.gamma_primary(normals['primary'][vis_test['primary']], join_vector),
             'secondary': re_numba.gamma_secondary(normals['secondary'][vis_test['secondary']], join_vector)}
        # negative sign is there because of reversed distance vector used for secondary component

        # testing mutual visibility of faces by assigning 0 to non visible face combination
        gamma['primary'][gamma['primary'] < 0] = 0.
        gamma['secondary'][gamma['secondary'] < 0] = 0.

        # calculating QAB = (cos gamma_a)*cos(gamma_b)/d**2
        q_ab = up.divide(up.multiply(gamma['primary'], gamma['secondary']), up.power(distance, 2))

        # calculating limb darkening factors for each combination of faces shape
        # (N_faces_primary * N_faces_secondary)

        # coefficients_primary = ld.interpolate_on_ld_grid()
        d_gamma = \
            {'primary': ld.limb_darkening_factor(coefficients=ldc['primary'][:, vis_test['primary']].T,
                                                 limb_darkening_law=settings.LIMB_DARKENING_LAW,
                                                 cos_theta=gamma['primary']),
             'secondary': ld.limb_darkening_factor(coefficients=ldc['secondary'][:, vis_test['secondary']].T,
                                                   limb_darkening_law=settings.LIMB_DARKENING_LAW,
                                                   cos_theta=gamma['secondary'].T).T
             }

        # precalculating matrix part of reflection effect correction
        matrix_to_sum2 = {cmp: q_ab * d_gamma[counterpart[cmp]] for cmp in components}

        for _ in range(iterations):
            for component in components:
                counterpart = settings.BINARY_COUNTERPARTS[component]

                # calculation of reflection effect correction as
                # 1 + (c / t_effi) * sum_j(r_j * Q_ab * t_effj^4 * D(gamma_j) * areas_j)
                # calculating vector part of reflection effect correction
                key_ = counterpart
                vector_to_sum1 = reflection_factor[key_] * teff4[key_][vis_test[key_]] * areas[key_][vis_test[key_]]
                counterpart_to_sum = up.matmul(vector_to_sum1, matrix_to_sum2['secondary']) \
                    if component == 'secondary' else up.matmul(matrix_to_sum2['primary'], vector_to_sum1)
                reflection_factor[component] = \
                    1 + (_c[component][vis_test[component]] / teff4[component][vis_test[component]]) * \
                    counterpart_to_sum

        for component in components:
            # assigning new temperatures according to last iteration as
            # teff_new = teff_old * reflection_factor^0.25
            temperatures[component][vis_test[component]] = \
                temperatures[component][vis_test[component]] * up.power(reflection_factor[component], 0.25)

    # redistributing temperatures back to the parent objects
    redistribute_temperatures(system, temperatures)
    return system


def build_temperature_distribution(system, components_distance, component="all"):
    """
    Function calculates temperature distribution across all surface faces.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param components_distance: float; distance of components in SMA units
    :param component: str; `primary` or `secondary`
    :return: system: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    if is_empty(component):
        logger.debug("no component set to build temperature distribution")
        return system

    components = bsutils.component_to_list(component)

    for component in components:
        star = getattr(system, component)

        logger.debug(f'computing effective temperature distribution '
                     f'on {component} component name: {star.name}')

        temperatures = btemperature.calculate_effective_temperatures(star, star.potential_gradient_magnitudes)
        setattr(star, "temperatures", temperatures)

        if star.has_spots():
            for spot_index, spot in star.spots.items():
                logger.debug(f'computing temperature distribution of spot {spot_index} / {component} component')

                pgms = spot.potential_gradient_magnitudes
                spot_temperatures = spot.temperature_factor * btemperature.calculate_effective_temperatures(star, pgms)
                setattr(spot, "temperatures", spot_temperatures)

        logger.debug(f'renormalizing temperature of components due to '
                     f'presence of spots in case of component {component}')
        renormalize_temperatures(star)

    if 'primary' in components and 'secondary' in components:
        logger.debug(f'calculating reflection effect with {settings.REFLECTION_EFFECT_ITERATIONS} iterations.')
        apply_reflection_effect(system, components_distance, settings.REFLECTION_EFFECT_ITERATIONS)
    return system


def init_surface_variables(star):
    """
    Function copies basic parameters of the stellar surface (points, faces, normals, temperatures, areas and log_g) of
    given star instance into new arrays during calculation of reflection effect.

    :param star: elisa.base.container.StarContainer;
    :return: Tuple; (points, faces, centres, normals, temperatures, areas)
    """
    points, faces = star.surface_serializer()
    centres = copy(star.face_centres)
    normals = copy(star.normals)
    temperatures = copy(star.temperatures)
    log_g = copy(star.log_g)
    areas = copy(star.areas)
    return points, faces, centres, normals, temperatures, areas, log_g


def include_spot_to_surface_variables(centres, spot_centres, normals, spot_normals, temperatures,
                                      spot_temperatures, areas, spot_areas, log_g, spot_log_g, vis_test, vis_test_spot):
    """
    Function includes surface parameters of spot faces into global arrays containing parameters from whole surface
    used in reflection effect.

    :param spot_log_g: numpy.array;
    :param log_g: numpy.array;
    :param centres: numpy.array;
    :param spot_centres: numpy.array; spot centres to append to `centres`
    :param normals: numpy.array;
    :param spot_normals: numpy.array; spot normals to append to `normals`
    :param temperatures: numpy.array;
    :param spot_temperatures: numpy.array; spot temperatures to append to `temperatures`
    :param areas: numpy.array;
    :param spot_areas: numpy.array; spot areas to append to `areas`
    :param vis_test: numpy.array;
    :param vis_test_spot: numpy.array; spot visibility test to append to `vis_test`
    :return: Tuple; (centres, normals, temperatures, areas, vis_test)
    """
    centres = np.append(centres, spot_centres, axis=0)
    normals = np.append(normals, spot_normals, axis=0)
    temperatures = np.append(temperatures, spot_temperatures, axis=0)
    areas = np.append(areas, spot_areas, axis=0)
    log_g = np.append(log_g, spot_log_g, axis=0)
    vis_test = np.append(vis_test, vis_test_spot, axis=0)

    return centres, normals, temperatures, areas, vis_test, log_g


def get_symmetrical_distance_matrix(shape, shape_reduced, centres, vis_test, vis_test_symmetry):
    """
    Function uses symmetries of the stellar component in order to reduce time in calculation distance matrix.

    :param shape: Tuple[int]; desired shape of join vector matrix
    :param shape_reduced: Tuple[int]; shape of the surface symmetries,
                         (faces above those indices are symmetrical to the ones below)
    :param centres: Dict;
    :param vis_test: Dict[str, numpy.array];
    :param vis_test_symmetry: Dict[str, numpy.array];
    :return: Tuple; (distance, join vector)

    ::

        distance - distance matrix
        join vector - matrix of unit vectors pointing between each two faces on opposite stars
    """
    # in case of symmetries, you need to calculate only minor part of distance matrix connected with base
    # symmetry part of the both surfaces
    distance = np.empty(shape=shape[:-1])
    join_vector = np.empty(shape=shape)

    distance[:shape_reduced[0], :], join_vector[:shape_reduced[0], :, :] = \
        utils.calculate_distance_matrix(
            points1=centres['primary'][vis_test_symmetry['primary']],
            points2=centres['secondary'][vis_test['secondary']],
            return_join_vector_matrix=True)

    aux = centres['primary'][vis_test['primary']]
    distance[shape_reduced[0]:, :shape_reduced[1]], join_vector[shape_reduced[0]:, :shape_reduced[1], :] = \
        utils.calculate_distance_matrix(points1=aux[shape_reduced[0]:],
                                        points2=centres['secondary'][vis_test_symmetry['secondary']],
                                        return_join_vector_matrix=True)

    return distance, join_vector


def get_symmetrical_gammma(shape, shape_reduced, normals, join_vector, vis_test, vis_test_symmetry):
    """
    Function uses surface symmetries to calculate cosine of angles between join vector and surface normals.

    :param shape: Tuple[int]; desired shape of gamma
    :param shape_reduced: Tuple[int]; shape of the surface symmetries, (faces above those
                                      indices are symmetrical to the ones below)
    :param normals: Dict[str, numpy.array];
    :param join_vector: Dict[str, numpy.array];
    :param vis_test: Dict[str, numpy.array];
    :param vis_test_symmetry: Dict[str, numpy.array];
    :return: gamma: Dict[str, numpy.array]; cos(angle(normal, join_vector))
    """
    gamma = {'primary': np.empty(shape=shape, dtype=FLOAT),
             'secondary': np.empty(shape=shape, dtype=FLOAT)}

    gamma['primary'][:, :shape_reduced[1]] = \
        re_numba.gamma_primary(normals['primary'][vis_test['primary']],
                               join_vector[:, :shape_reduced[1], :])
    gamma['primary'][:shape_reduced[0], shape_reduced[1]:] = \
        re_numba.gamma_primary(normals['primary'][vis_test_symmetry['primary']],
                               join_vector[:shape_reduced[0], shape_reduced[1]:, :])

    gamma['secondary'][:shape_reduced[0], :] = \
        re_numba.gamma_secondary(normals['secondary'][vis_test['secondary']],
                                 join_vector[:shape_reduced[0], :, :])
    gamma['secondary'][shape_reduced[0]:, :shape_reduced[1]] = \
        re_numba.gamma_secondary(normals['secondary'][vis_test_symmetry['secondary']],
                                 join_vector[shape_reduced[0]:, :shape_reduced[1], :])

    return gamma


def check_symmetric_gamma_for_negative_num(gamma, shape_reduced):
    """
    If cos < 0 it will be redefined as 0 are inplaced.

    :param gamma: Dict[str, numpy.array];
    :param shape_reduced: Tuple[int];
    """
    gamma['primary'][:, :shape_reduced[1]][gamma['primary'][:, :shape_reduced[1]] < 0] = 0.
    gamma['primary'][:shape_reduced[0], shape_reduced[1]:][gamma['primary'][:shape_reduced[0],
                                                           shape_reduced[1]:] < 0] = 0.
    gamma['secondary'][:shape_reduced[0], :][gamma['secondary'][:shape_reduced[0], :] < 0] = 0.
    gamma['secondary'][shape_reduced[0]:, :shape_reduced[1]][gamma['secondary'][shape_reduced[0]:,
                                                             :shape_reduced[1]] < 0] = 0.


def get_symmetrical_d_gamma(shape, shape_reduced, ldc, gamma):
    """
    Function uses surface symmetries to calculate limb darkening factor matrices
    for each component that are used in reflection effect.

    :param ldc: Dict; arrays of limb darkening coefficients for each face of each component
    :param shape: Tuple; desired shape of limb darkening matrices d_gamma
    :param shape_reduced: Tuple; shape of the surface symmetries, (faces above those
                                 indices are symmetrical to the ones below)
    :param gamma: Dict; cosines of angles between join vector and surface element normal vector for both components
    :return: Dict; limb darkening coefficient matrix
    """
    d_gamma = {'primary': np.empty(shape=shape, dtype=FLOAT),
               'secondary': np.empty(shape=shape, dtype=FLOAT)}

    cos_theta = gamma['primary'][:, :shape_reduced[1]]
    d_gamma['primary'][:, :shape_reduced[1]] = ld.limb_darkening_factor(
        coefficients=ldc['primary'][:, :shape[0]].T,
        limb_darkening_law=settings.LIMB_DARKENING_LAW,
        cos_theta=cos_theta)

    cos_theta = gamma['primary'][:shape_reduced[0], shape_reduced[1]:]
    d_gamma['primary'][:shape_reduced[0], shape_reduced[1]:] = ld.limb_darkening_factor(
        coefficients=ldc['primary'][:, :shape_reduced[0]].T,
        limb_darkening_law=settings.LIMB_DARKENING_LAW,
        cos_theta=cos_theta)

    cos_theta = gamma['secondary'][:shape_reduced[0], :]
    d_gamma['secondary'][:shape_reduced[0], :] = ld.limb_darkening_factor(
        coefficients=ldc['secondary'][:, :shape[1]].T,
        limb_darkening_law=settings.LIMB_DARKENING_LAW,
        cos_theta=cos_theta.T).T

    cos_theta = gamma['secondary'][shape_reduced[0]:, :shape_reduced[1]]
    d_gamma['secondary'][shape_reduced[0]:, :shape_reduced[1]] = ld.limb_darkening_factor(
        coefficients=ldc['secondary'][:, :shape_reduced[1]].T,
        limb_darkening_law=settings.LIMB_DARKENING_LAW,
        cos_theta=cos_theta.T).T

    return d_gamma


def get_symmetrical_q_ab(shape, shape_reduced, gamma, distance):
    """
    Function uses surface symmetries to calculate parameter::

        QAB = (cos gamma_a)*cos(gamma_b)/d**2

    in reflection effect.

    :param shape: Tuple[int]; desired shape of q_ab
    :param shape_reduced: Tuple[int]; shape of the surface symmetries,
                                     (faces above those indices are symmetrical to the ones below)
    :param gamma: Dict[str, numpy.array];
    :param distance: numpy.array;
    :return: numpy.array;
    """
    q_ab = np.empty(shape=shape, dtype=FLOAT)
    q_ab[:, :shape_reduced[1]] = \
        up.divide(up.multiply(gamma['primary'][:, :shape_reduced[1]],
                              gamma['secondary'][:, :shape_reduced[1]]),
                  up.power(distance[:, :shape_reduced[1]], 2))
    q_ab[:shape_reduced[0], shape_reduced[1]:] = \
        up.divide(up.multiply(gamma['primary'][:shape_reduced[0], shape_reduced[1]:],
                              gamma['secondary'][:shape_reduced[0], shape_reduced[1]:]),
                  up.power(distance[:shape_reduced[0], shape_reduced[1]:], 2))
    return q_ab


def get_distance_matrix_shape(system, vis_test):
    """
    Calculates shapes of distance and join vector matrices along with shapes
    of symmetrical parts of those matrices used in reflection effect.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param vis_test: numpy.array;
    :return: Tuple;
    """
    shape = (np.sum(vis_test['primary']), np.sum(vis_test['secondary']), 3)
    shape_reduced = (np.sum(system.primary.symmetry_faces(vis_test['primary'])),
                     np.sum(system.secondary.symmetry_faces(vis_test['secondary'])))
    return shape, shape_reduced


def interpolate_albedo(temperature):
    """
    Quick interpolator for determination of default value of albedo. Interpolation data are from Figure 6 in
    Claret 2001 MNRAS 327, 989–994.

    :param temperature: float
    :return: float; interpolated albedo (0, 1)
    """
    with open(settings.PATH_TO_ALBEDOS, 'r') as f:
        data = json.load(f)
    interp_temps = data['x']
    interp_a = data['y']

    if temperature <= 0:
        raise ValueError('Negative temperature of the star encountered.')

    return np.interp(np.log10(temperature), interp_temps, interp_a)



import numpy as np

from . import utils
from ... single_system import (
    utils as ssutils,
)


def produce_curves_wo_pulsations_mp(*args):
    """
    Calculates curve for single system without pulsations (surface geometry is not changed).

    :param args: Tuple;
    :**args options**:
        * ** single_system ** * - elisa.single_system.system.SingleSystem
        * ** system_container ** * - elisa.single_system.container.SystemContainer
        * ** phases ** * - array; phases in which to calculate curves
        * ** crv_labels ** * - List;
        * ** curves_fn ** * - function to calculate curve points at given orbital positions
        * ** kwargs ** * - Dict;

    :return:
    """
    single, initial_system, phase_batch, crv_labels, curves_fn, kwargs = args
    position_method = kwargs.pop("position_method")

    rotational_motion = position_method(input_argument=phase_batch, return_nparray=False, calculate_from='phase')
    curves = {key: np.zeros(phase_batch.shape) for key in crv_labels}

    for pos_idx, position in enumerate(rotational_motion):
        on_pos = ssutils.move_sys_onpos(initial_system, position)
        star = on_pos.star

        setattr(star, 'coverage', star.areas)

        curves = curves_fn(curves, pos_idx, crv_labels, on_pos)

    return curves


def produce_curves_with_pulsations_mp(*args):
    """
    Calculates curve for single system with pulsations.

    :param args:
    :**args options**:
        * ** single_system ** * - elisa.single_system.system.SingleSystem
        * ** system_container ** * - elisa.single_system.container.SystemContainer
        * ** phases ** * - array; phases in which to calculate curves
        * ** crv_labels ** * - List;
        * ** curves_fn ** * - function to calculate curve points at given orbital positions
        * ** kwargs ** * - Dict;

    :return:
    """
    single, initial_system, phase_batch, crv_labels, curves_fn, kwargs = args
    position_method = kwargs.pop("position_method")

    rotational_motion = position_method(input_argument=phase_batch, return_nparray=False, calculate_from='phase')
    curves = {key: np.zeros(phase_batch.shape) for key in crv_labels}

    for pos_idx, position in enumerate(rotational_motion):
        sys_on_pos = initial_system.copy()
        sys_on_pos.set_on_position_params(position)
        sys_on_pos.set_time()

        sys_on_pos.build_pulsations()
        utils.prep_surface_params(sys_on_pos, return_values=False, write_to_containers=True, **kwargs)

        sys_on_pos = ssutils.move_sys_onpos(sys_on_pos, position, on_copy=False)
        star = sys_on_pos.star
        setattr(star, 'coverage', star.areas)

        curves = curves_fn(curves, pos_idx, crv_labels, sys_on_pos)

    return curves



import numpy as np


def set_axes_equal(ax):
    """
    Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    """

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5*max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])



import numpy as np

from .. import utils
from .. base.body import Body
from .. base.container import StarPropertiesContainer
from .. base.transform import StarProperties
from .. pulse.mode import PulsationMode
from .. logger import getLogger
from .. import units as u

from copy import (
    copy,
    deepcopy
)

logger = getLogger('base.star')


class Star(Body):
    """
    Child class of elisa.base.body.Body representing `Star`.
    
    Class can be imported directly:
    ::

        from elisa import Star
    
    Mandatory `Star` arguments if the instance is a component of the `SingleSystem`:

        :param mass: float; If mass is int, np.int, float, np.float, program assumes solar mass as it's unit.
                        If mass astropy.unit.quantity.Quantity instance, program converts it to default units.
        :param t_eff: float; Accepts value in any temperature unit. If your input is without unit,
                         function assumes that supplied value is in K.
        :param polar_log_g: float; log_10 of the polar surface gravity

    following optional arguments are also available:

        :param metallicity: float; log[M/H] default value is 0.0
        :param gravity_darkening: float; gravity darkening factor, if not supplied, it is interpolated from Claret 2003
                                         based on t_eff
        :param limb_darkening_coefficients: Union[float, dict]; optional limb darkening coefficients
                                            used for the whole star useful in case the modelled star is outside the
                                            supported range of atmospheric parameters. Limb darkening coefficients can
                                            be supplied as dict {passband: ld_coefs}. If unused, elisa will
                                            interpolate the values from supplied limb-darkening tables.

    After initialization of the SingleSystem, following additional attributes of the Star instance are available:

        :critical_potential: float; potential of the star required to fill its Roche lobe
        :equivalent_radius: float; radius of a sphere with the same volume as a component (in SMA units)
        :polar_radius: float; radius of a star towards the pole of the star
        :equatorial_radius: float; radius of a star towards the pole of the star
    
    Mandatory `Star` arguments if the instance is a component of the `BinarySystem`:

        :param mass: float; If mass is int, np.int, float, np.float, program assumes solar mass as it's unit.
                            If mass astropy.unit.quantity.Quantity instance, program converts it to default units.
        :param t_eff: float; Accepts value in any temperature unit. If your input is without unit,
                             function assumes that supplied value is in K.
        :param surface_potential: float; generalized surface potential (Wilson 79)
        :param synchronicity: float; synchronicity F (omega_rot / omega_orb), equals 1 for synchronous rotation
        :param albedo: float; surface albedo, value from <0, 1> interval

    following mandatory arguments are also available:

        :param metallicity: float; log[M/H] default value is 0.0
        :param gravity_darkening: float; gravity darkening factor, if not supplied, it is interpolated from Claret 2003
                                         based on t_eff


    After initialization of the `BinarySystem`, following additional attributes of the `Star` instance are available:

        :critical_potential: float; potential of the star required to fill its Roche lobe
        :equivalent_radius: float; radius of a sphere with the same volume as a component (in SMA units)
        :filling_factor: float: calculated as (Omega_{inner} - Omega) / (Omega_{inner} - Omega_{outter})

                            :filling factor < 0: component does not fill its Roche lobe
                            :filling factor = 0: component fills preciselly its Roche lobe
                            :1 > filling factor > 0: component overflows its Roche lobe
                            :filling factor = 1: upper boundary of the filling factor, higher value would lead to 
                                                 the mass loss trough Lagrange point L2
                                                 
        Radii at periastron (in SMA units)
            :polar_radius: float; radius of a star towards the pole of the star
            :side_radius: float; radius of a star in the direction perpendicular to the pole and direction of a
                                 companion
            :backward_radius: float; radius of a star in the opposite direction as the binary companion
            :forward_radius: float; radius of a star towards the binary companion, returns numpy.nan if the system is
                                    over-contact

    Optional parameters of `Star` instances can be defined with the following arguments:
        :param spots: List[Dict[str, float]]; Spots definitions. Order in which the spots are defined will determine the
                                          layering of the spots (spot defined as first will lay bellow any subsequently
                                          defined overlapping spot). Example of spots definition:

        ::

            [
                 {"longitude": 90,
                  "latitude": 58,
                  "angular_radius": 15,
                  "temperature_factor": 0.9},
                 {"longitude": 85,
                  "latitude": 80,
                  "angular_radius": 30,
                  "temperature_factor": 1.05},
                 {"longitude": 45,
                  "latitude": 90,
                  "angular_radius": 30,
                  "temperature_factor": 0.95},
             ]

        :param pulsations: List[Dict[str, float]]; to be added soon
        :param atmosphere: str; atmosphere to use for given object instance, available atmosphere models:

            - `castelli`, `castelli-kurucz`, `ck` or `ck04`: atmosphere models in Castelli-Kurucz, 2004
            - `kurucz`, `k` or `k93`: atmosphere models in Kurucz, 1993

    After initialization, `Star` instance after initialization within the given `System` has its spot initialized in a
    `star_instance.spots` attribute as a list containing elisa.base.spot.Spot containers.

    """

    MANDATORY_KWARGS = ['mass', 't_eff']
    OPTIONAL_KWARGS = ['surface_potential', 'synchronicity', 'albedo', 'pulsations', 'atmosphere',
                       'spots', 'metallicity', 'polar_log_g', 'discretization_factor', 'gravity_darkening',
                       'limb_darkening_coefficients']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, name=None, **kwargs):
        utils.invalid_kwarg_checker(kwargs, Star.ALL_KWARGS, Star)
        super(Star, self).__init__(name, **kwargs)
        kwargs = self.transform_input(**kwargs)

        # default values of properties
        self.filling_factor = np.nan
        self.critical_surface_potential = np.nan
        self.surface_potential = np.nan
        self.metallicity = 0.0
        self.polar_log_g = np.nan
        self.gravity_darkening = np.nan
        self.limb_darkening_coefficients = None
        self._pulsations = list()

        self.side_radius = np.nan
        self.forward_radius = np.nan
        self.backward_radius = np.nan
        self.equivalent_radius = np.nan

        self.init_parameters(**kwargs)

    @property
    def default_input_units(self):
        """
        Returns set of default units of intialization parameters, in case, when provided without units.

        :return: elisa.units.DefaultStarInputUnits;
        """
        return u.DefaultStarInputUnits

    @property
    def default_internal_units(self):
        """
        Returns set of internal units of Star parameters.

        :return: elisa.units.DefaultStarUnits;
        """
        return u.DefaultStarUnits

    def transform_input(self, **kwargs):
        """
        Transform and validate input kwargs.

        :param kwargs: Dict;
        :return: Dict;
        """
        return StarProperties.transform_input(**kwargs)

    def init_parameters(self, **kwargs):
        """
        Initialise instance parameters

        :param kwargs: Dict; initial parameters
        :return:
        """
        logger.debug(f"initialising properties of class instance {self.__class__.__name__}")
        for kwarg in Star.ALL_KWARGS:
            if kwarg in kwargs:
                setattr(self, kwarg, kwargs[kwarg])

    def kwargs_serializer(self):
        """
        Creating dictionary of keyword arguments in order to be able to
        reinitialize the class instance in init().

        :return: Dict;
        """
        default_units = {
            "mass": u.kg,
            "t_eff": u.K,
            "discretization_factor": u.rad
        }

        serialized_kwargs = dict()
        for kwarg in self.ALL_KWARGS:
            if kwarg in ["spots"]:
                # important: this relies on dict ordering
                value = [spot.kwargs_serializer() for spot in getattr(self, kwarg).values()]
            elif kwarg in default_units:
                value = getattr(self, kwarg)
                if not isinstance(value, u.Quantity):
                    value = value * default_units[kwarg]
            else:
                value = getattr(self, kwarg)

            serialized_kwargs[kwarg] = value
        return serialized_kwargs

    def init(self):
        self.__init__(**self.kwargs_serializer())

    def has_pulsations(self):
        """
        Determine whether Star has defined pulsations.

        :return: bool;
        """
        return len(self._pulsations) > 0

    @property
    def pulsations(self):
        """
        Return pulsation modes for given Star instance.

        :return: Dict;

        ::

            {index: PulsationMode}
        """
        return self._pulsations

    @pulsations.setter
    def pulsations(self, pulsations):
        """
        Set pulsation mode for given Star instance defined by dict.

        :param pulsations: Dict;

        ::

            [{"l": <int>, "m": <int>, "amplitude": <float>, "frequency": <float>}, ...]

        :return:
        """
        if pulsations in [None, []]:
            self._pulsations = {}
        elif pulsations:
            self._pulsations = {idx: PulsationMode(**pulsation_meta) for idx, pulsation_meta in enumerate(pulsations)}

    def properties_serializer(self):
        """
        Prepares properties to be inherited from Star instance by StarContainer instance.

        :return: Dict;
        """
        properties_list = ['mass', 't_eff', 'synchronicity', 'albedo', 'discretization_factor', 'equivalent_radius',
                           'polar_radius', 'equatorial_radius', 'gravity_darkening', 'surface_potential', 'pulsations',
                           'metallicity', 'polar_log_g', 'critical_surface_potential', 'atmosphere',
                           # todo: remove side_radius when figured out starting point for solver
                           'side_radius', 'limb_darkening_coefficients']
        props = {prop: copy(getattr(self, prop)) for prop in properties_list}
        props.update({
            "name": self.name,
            "spots": deepcopy(self.spots)
        })
        return props

    def to_properties_container(self):
        """
        Serialize instance of elisa.base.star.Star to elisa.base.container.StarPropertiesContainer.

        :return: elisa.base.container.StarPropertiesContainer
        """
        return StarPropertiesContainer(**self.properties_serializer())



